
(******************************************************************************)
(*                         DFT Gates and Operators                            *)
(*                      Definitions and Probabilities                         *)
(*                       and Simplification Theorems                          *)
(*                                  by                                        *)
(*                          Yassmeen Elderhalli                               *)
(*                      y_elderh@encs.concordia.ca                            *)
(******************************************************************************)

(*loadPath := "/home/yassmeen/Downloads/ProbTheories_HOL-k12" :: !loadPath;
loadPath := "/home/yassmeen/Downloads/My-Theories" :: !loadPath;

*)
(*set_trace "Unicode" 0;*)



(*app load ["HolKernel", "Parse", "boolLib", "bossLib", "numLib", "unwindLib", 
"tautLib", "Arith", "prim_recTheory", "combinTheory", "quotientTheory", 
"arithmeticTheory", "realTheory", "hrealTheory", "realaxTheory", "realLib",  
"jrhUtils", "pairTheory", "seqTheory", "limTheory", "transcTheory", "listTheory", 
"mesonLib", "boolTheory", "topologyTheory", "pred_setTheory", "util_probTheory", 
"optionTheory", "numTheory", "sumTheory", "InductiveDefinition", "ind_typeTheory",
"pred_setLib", "iterate_hvgTheory", "card_hvgTheory", "product_hvgTheory", 
"topology_hvgTheory", "derivative_hvgTheory", "integration_hvgTheory",
"real_sigmaTheory", "extreal_hvgTheory", "measure_hvgTheory", "lebesgue_hvgTheory",
"probability_hvgTheory", "lebesgue_measure_hvgTheory", "normal_rv_hvgTheory", "netsTheory",
"PIE_EXTREALTheory", "integration_before_afterTheory", "dep_rewrite"];
*)
open HolKernel Parse boolLib bossLib numLib unwindLib tautLib Arith prim_recTheory 
combinTheory quotientTheory arithmeticTheory hrealTheory realaxTheory realTheory 
realLib jrhUtils pairTheory seqTheory limTheory transcTheory listTheory mesonLib 
boolTheory topologyTheory pred_setTheory util_probTheory optionTheory numTheory 
sumTheory InductiveDefinition ind_typeTheory pred_setLib iterate_hvgTheory 
card_hvgTheory product_hvgTheory topology_hvgTheory derivative_hvgTheory
integration_hvgTheory real_sigmaTheory extreal_hvgTheory measure_hvgTheory
lebesgue_hvgTheory probability_hvgTheory lebesgue_measure_hvgTheory
normal_rv_hvgTheory netsTheory PIE_EXTREALTheory integration_before_afterTheory dep_rewrite;


val _ = new_theory "DFT_Gates_def_PROB";

infixr 0 ++ << || ORELSEC ## --> THENC;
infix 1 >> |->;
fun parse_with_goal t (asms, g) =
  let
    val ctxt = free_varsl (g::asms)
  in
    Parse.parse_in_context ctxt t
  end;
 
val PARSE_TAC = fn tac => fn q => W (tac o parse_with_goal q);
val Suff = PARSE_TAC SUFF_TAC;
val POP_ORW = POP_ASSUM (fn thm => ONCE_REWRITE_TAC [thm]);
val !! = REPEAT;
val op++ = op THEN;
val op<< = op THENL;
val op|| = op ORELSE;
val op>> = op THEN1;
val std_ss' = simpLib.++ (std_ss, boolSimps.ETA_ss);

(*---------------------------*)
fun SET_TAC L =
    POP_ASSUM_LIST(K ALL_TAC) THEN REPEAT COND_CASES_TAC THEN
    REWRITE_TAC (append [EXTENSION, SUBSET_DEF, PSUBSET_DEF, DISJOINT_DEF,
    SING_DEF] L) THEN SIMP_TAC std_ss [NOT_IN_EMPTY, IN_UNIV, IN_UNION,
    IN_INTER, IN_DIFF, IN_INSERT, IN_DELETE, IN_BIGINTER, IN_BIGUNION,
    IN_IMAGE, GSPECIFICATION, IN_DEF] THEN METIS_TAC [];

fun SET_RULE tm = prove(tm,SET_TAC []);


val set_rewrs
= [INTER_EMPTY, INTER_UNIV, UNION_EMPTY, UNION_UNIV, DISJOINT_UNION,
DISJOINT_INSERT, DISJOINT_EMPTY, GSYM DISJOINT_EMPTY_REFL,
DISJOINT_BIGUNION, INTER_SUBSET, INTER_IDEMPOT, UNION_IDEMPOT,
SUBSET_UNION];

val UNIONL_def = Define `(UNIONL [] = {})
/\ (UNIONL (s::ss) = (s:'a ->bool) UNION UNIONL ss)`;


val IN_UNIONL = store_thm
("IN_UNIONL",
``!l (v:'a ). v IN UNIONL l = (?s. MEM s l /\ v IN s)``,
Induct >> RW_TAC std_ss [UNIONL_def, MEM, NOT_IN_EMPTY]
++ RW_TAC std_ss [UNIONL_def, MEM, NOT_IN_EMPTY, IN_UNION]
++ PROVE_TAC []);


val elt_rewrs
= [SUBSET_DEF, IN_COMPL, IN_DIFF, IN_UNIV, NOT_IN_EMPTY, IN_UNION,
IN_INTER, IN_IMAGE, IN_FUNSET, IN_DFUNSET, GSPECIFICATION,
DISJOINT_DEF, IN_BIGUNION, IN_BIGINTER, IN_COUNT, IN_o,
IN_UNIONL, IN_DELETE, IN_PREIMAGE, IN_SING, IN_INSERT];


fun rewr_ss ths =
simpLib.++
(std_ss,
simpLib.SSFRAG
{ac = [],
name = NONE,
convs = [],
dprocs = [],
filter = NONE,
rewrs = set_rewrs @ elt_rewrs,
congs = []});
val pset_set_ss = rewr_ss set_rewrs;
val pset_elt_ss = rewr_ss elt_rewrs;
val pset_ss = rewr_ss (set_rewrs @ elt_rewrs);


fun PSET_TAC ths =
REPEAT (POP_ASSUM MP_TAC)
++ RW_TAC pset_set_ss ths
++ REPEAT (POP_ASSUM MP_TAC)
++ RW_TAC pset_elt_ss ths;

fun parse_with_goal t (asms, g) =
  let
    val ctxt = free_varsl (g::asms)
  in
    Parse.parse_in_context ctxt t
  end;
val PARSE_TAC = fn tac => fn q => W (tac o parse_with_goal q);
val Simplify = RW_TAC arith_ss;
val Suff = PARSE_TAC SUFF_TAC;
val Know = PARSE_TAC KNOW_TAC;
fun wrap a = [a];
val Rewr = DISCH_THEN (REWRITE_TAC o wrap);
val Rewr' = DISCH_THEN (ONCE_REWRITE_TAC o wrap);
val Cond =
  DISCH_THEN
  (fn mp_th =>
   let
     val cond = fst (dest_imp (concl mp_th))
   in
     KNOW_TAC cond << [ALL_TAC, DISCH_THEN (MP_TAC o MP mp_th)]
   end);
val FUN_EQ_TAC = CONV_TAC (CHANGED_CONV (ONCE_DEPTH_CONV FUN_EQ_CONV));
fun K_TAC _ = ALL_TAC;
fun MESON ths tm = prove(tm,MESON_TAC ths);
fun METIS ths tm = prove(tm,METIS_TAC ths);
val DISC_RW_KILL = DISCH_TAC THEN ONCE_ASM_REWRITE_TAC [] THEN 
                   POP_ASSUM K_TAC;
			   
val IN_REST = store_thm ("IN_REST",
 ``!x:'a. !s. x IN (REST s) <=> x IN s /\ ~(x = CHOICE s)``,
  REWRITE_TAC[REST_DEF, IN_DELETE]);
fun SET_TAC L = 
    POP_ASSUM_LIST(K ALL_TAC) THEN REPEAT COND_CASES_TAC THEN
    REWRITE_TAC (append [EXTENSION, SUBSET_DEF, PSUBSET_DEF, DISJOINT_DEF,
    SING_DEF] L) THEN
    SIMP_TAC std_ss [NOT_IN_EMPTY, IN_UNIV, IN_UNION, IN_INTER, IN_DIFF, 
      IN_INSERT, IN_DELETE, IN_REST, IN_BIGINTER, IN_BIGUNION, IN_IMAGE, 
      GSPECIFICATION, IN_DEF, EXISTS_PROD] THEN METIS_TAC [];
fun ASSERT_TAC tm = SUBGOAL_THEN tm STRIP_ASSUME_TAC;
fun SET_RULE tm = prove(tm,SET_TAC []);
fun ASM_SET_TAC L = REPEAT (POP_ASSUM MP_TAC) THEN SET_TAC L;
val ASM_ARITH_TAC = REPEAT (POP_ASSUM MP_TAC) THEN ARITH_TAC;
val ASM_REAL_ARITH_TAC = REPEAT (POP_ASSUM MP_TAC) THEN REAL_ARITH_TAC;



(******************************************************************************)
(*                                                                            *)
(*                     DFT gates and Operators Definitions                    *)
(*                                                                            *)
(******************************************************************************)

val DFT_event_def = Define ` 
          DFT_event  p (X:'a->extreal) t  =  {s|  X s  <=Normal t } INTER p_space p`;

val D_OR_def = Define ` D_OR (X: 'a->extreal) Y   = (\s. min  (X s) (Y s))`;

val D_AND_def = Define ` D_AND (X: 'a->extreal) Y   = (\s. max (X s)( Y s))`;

val D_SIMULT_def = Define ` 
     D_SIMULT (X: 'a->extreal) Y   = (\s. if  ((X s) = ( Y s )) then (X s) else PosInf)`;

val D_BEFORE_def = Define `
     D_BEFORE (X: 'a->extreal) Y   = (\s. if  (X s) < (Y s) then (X s) else PosInf)`;

val D_INCLUSIVE_BEFORE_def = Define ` 
     D_INCLUSIVE_BEFORE (X:'a->extreal) Y   = (\s. if  (X s) <=(Y s) then (X s) else PosInf)`;

val NEVER_def = Define ` NEVER = (\s. PosInf)`;

val ALWAYS_def = Define ` ALWAYS = (\s. 0:extreal)`;

val P_AND_def = Define ` 
    P_AND (X: 'a->extreal) Y  =  (\s. if  (X s) <= (Y s) then (Y s) else PosInf)`;

val FDEP_def = Define ` FDEP (X:'a-> extreal)  Tr   =(\s.  min  (X s)  (Tr s)) ` ;

val CSP_def = Define ` CSP (Y:'a-> extreal) X =( \s. if (Y s < X s) then  X s else PosInf)`;

val WSP_def = Define ` 
    WSP (Y:'a->extreal) X_a X_d = 
     D_OR
       (D_OR
          (D_OR (D_AND X_a (D_BEFORE Y X_a)) (D_AND Y (D_BEFORE X_d Y)))
          (D_SIMULT Y X_a)) (D_SIMULT Y  X_d)`;

val HSP_def = Define ` HSP (Y:'a->extreal) X  =(\s.  max (Y s )  (X s)) `;

val Never_events_def = Define `  NEVER_events X Y = (D_AND X Y = NEVER)`;

val shared_spare_def = Define `
     shared_spare X Y Z_a Z_d =  
     D_OR (D_OR (D_AND X (D_BEFORE Z_d X)) (D_AND X (D_BEFORE Y X)))
       (D_AND Z_a (D_BEFORE X Z_a))`;


val PAND_DEF2 = store_thm("PAND_DEF2",
  ``  ! X Y. P_AND X Y = D_AND Y (D_INCLUSIVE_BEFORE X Y)``,

REPEAT GEN_TAC
THEN REWRITE_TAC [D_AND_def, D_INCLUSIVE_BEFORE_def, P_AND_def]
THEN REWRITE_TAC [extreal_max_def]
THEN RW_TAC std_ss[]
THEN KNOW_TAC (``
  !s.
  (if X s <= Y s then Y s else PosInf) =
  if Y s <= if X s <= Y s then X s else PosInf then
    if X s <= Y s then X s else PosInf
  else Y s``) 
    THEN1(RW_TAC std_ss[]
          THEN1( METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
    THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);


val FDEP_OR = store_thm("FDEP_OR",
  `` ! X Y. FDEP X Y = D_OR X Y``,
RW_TAC std_ss[D_OR_def, FDEP_def]);

val HSP_AND = store_thm("HSP_AND",
  ``  ! X Y. HSP Y X  = D_AND Y X``,
  RW_TAC std_ss [HSP_def, D_AND_def]);


val WSP_HSP = store_thm("WSP_HSP",
  `` !X Y. WSP Y X X = HSP Y X ``,
  RW_TAC std_ss [WSP_def, HSP_def, D_OR_def, D_AND_def, D_SIMULT_def, D_BEFORE_def]
  THEN KNOW_TAC(``!s.(
        min
          (min
             (min (max (X s) (if Y s < X s then Y s else PosInf))
                (max (Y s) (if X s < Y s then X s else PosInf)))
             (if Y s = X s then X s else PosInf))
          (if Y s = X s then X s else PosInf)) =
   ( max (Y s) (X s)) ``)

       THEN1(RW_TAC std_ss[extreal_max_def, extreal_min_def] THEN REPEAT( POP_ASSUM MP_TAC)
THEN  RW_TAC std_ss[] THEN METIS_TAC [le_lt, lt_le, lt_infty, le_infty, extreal_lt_def, extreal_le_def, lte_trans, let_trans])
  THEN RW_TAC std_ss[]);


val WSP_CSP = store_thm("WSP_CSP",

`` !X_a X_d Y. (!s. ALL_DISTINCT [X_a s; Y s]) /\ (X_d = NEVER) ==>
               (WSP Y X_a X_d = CSP Y X_a)``,

   RW_TAC std_ss[WSP_def, CSP_def, D_OR_def, D_AND_def, D_SIMULT_def, D_BEFORE_def]
     THEN KNOW_TAC(`` !s.
        (min
          (min
             (min (max (X_a s) (if Y s < X_a s then Y s else PosInf))
                (max (Y s) (if NEVER s < Y s then NEVER s else PosInf)))
             (if Y s = X_a s then X_a s else PosInf))
          (if Y s = NEVER s then NEVER s else PosInf)) =
   ( if Y s < X_a s then X_a s else PosInf) ``)
       THEN1(RW_TAC std_ss[extreal_max_def, extreal_min_def, NEVER_def] THEN REPEAT( POP_ASSUM MP_TAC)
THEN  RW_TAC std_ss[] THEN FULL_SIMP_TAC std_ss[ALL_DISTINCT, MEM] THEN METIS_TAC [le_lt, lt_le, lt_infty, le_infty, extreal_lt_def, extreal_le_def, lte_trans, let_trans, ALL_DISTINCT, MEM])
    THEN RW_TAC std_ss[]);

(* -------------------------------------------------------------------------- *)
(*                                                                            *)
(*                Probability of Gates and Operators                          *)
(*                                                                            *)
(* -------------------------------------------------------------------------- *)


val AND_INTER = store_thm("AND_INTER",
  `` !p t X Y.  DFT_event p (D_AND X Y )t =(DFT_event p X t)  INTER (DFT_event p Y t )``,
RW_TAC std_ss [DFT_event_def, D_AND_def] 
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN METIS_TAC [max_le]);

val OR_UNION = store_thm("OR_UNION",
  ``  !p t X Y.  DFT_event p (D_OR X Y ) t = (DFT_event p X t) UNION (DFT_event p Y t)``,

RW_TAC std_ss [DFT_event_def, D_OR_def] 
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN METIS_TAC [min_le]);

val lborel_le = store_thm("lborel_le",
  `` ! x:real. {w|w<=x} IN measurable_sets lborel``,
RW_TAC std_ss[]
THEN RW_TAC std_ss [borel_closed, lborel, measurable_sets_def, subsets_def,CLOSED_HALFSPACE_COMPONENT_LE ]);

val DFT_PREIMAGE = store_thm("DFT_PREIMAGE",
  ``  ! X p t. DFT_event p X t = PREIMAGE X {u|u<= Normal t} INTER p_space p``,

RW_TAC std_ss [DFT_event_def] THEN
SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]);

val PREIMAGE_EXTREAL_REAL = store_thm("PREIMAGE_EXTREAL_REAL",
  `` ! X t p. (!s. X s <> PosInf /\ 0<= X s) ==>
        (PREIMAGE X {u|u<= Normal t} INTER p_space p =
	 PREIMAGE (\s. real (X s)) {u|u<= t:real} INTER p_space p)``, 
RW_TAC std_ss []
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]
THEN EQ_TAC THEN1(RW_TAC std_ss[]
     	    THEN RW_TAC std_ss [real_def]
	    	 THEN1( METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
            THEN SELECT_ELIM_TAC
	    THEN RW_TAC std_ss[]
	    	 THEN1(Q.EXISTS_TAC` real (X x)`
		       THEN METIS_TAC [normal_real])
		 THEN METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
THEN RW_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [real_def]
THEN UNDISCH_TAC(``(if (X x = NegInf) \/ (X x = PosInf) then 0
       else (@r. X x = Normal r)) <= t``)
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[] THEN1 (Q.EXISTS_TAC` real (X x)`
     	    	            THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def]);

val PREIMAGE_EXTREAL_REAL_2RV = store_thm("PREIMAGE_EXTREAL_REAL_2RV",
  ``!X Y t p.
       (!s. X s <> PosInf /\ 0 <= X s /\ Y s <> PosInf /\ 0 <= Y s) /\
       0 <= t ==>
       (PREIMAGE (\s. (X s,Y s))
          {(u,w) | u < w /\ 0 <= w /\ w <= Normal t} INTER p_space p =
        PREIMAGE (\s. (real( X s),real (Y s)))
          {(u,w) | u < w /\ 0 <= w /\ w <= t} INTER p_space p)``, 

RW_TAC std_ss [PREIMAGE_def]
THEN KNOW_TAC(`` !s. (X s,Y s) IN {(u,w) | u < w /\ 0<=w/\w <= Normal t} =
                       X s < Y s /\ 0<= Y s/\ Y s <= Normal t``)
THEN1(RW_TAC std_ss [IN_ELIM_PAIR_THM])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN KNOW_TAC(`` !s. (real (X s),real(Y s)) IN {(u,w) | u < w /\ 0<=w/\w <= t} =
                       real(X s) < real(Y s) /\ 0<= real (Y s) /\ real(Y s) <= t``)
THEN1(RW_TAC std_ss [IN_ELIM_PAIR_THM])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN EQ_TAC
THEN1 (RW_TAC std_ss[]
      THEN1 (KNOW_TAC (``X x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
      	    THEN DISCH_TAC
      	    THEN KNOW_TAC (``Y x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
      	    THEN DISCH_TAC
      	    THEN METIS_TAC [extreal_real_lt])
      THEN1(RW_TAC std_ss[real_def]
      THEN RW_TAC std_ss [REAL_LE_REFL]
      THEN SELECT_ELIM_TAC
      THEN RW_TAC std_ss []
      THEN1 (Q.EXISTS_TAC `real (Y x)`
      	    THEN METIS_TAC [normal_real])
      THEN METIS_TAC [extreal_le_eq, extreal_of_num_def])
      THEN RW_TAC std_ss[real_def]
      THEN SELECT_ELIM_TAC
      THEN RW_TAC std_ss []
      THEN1 (Q.EXISTS_TAC `real (Y x)`
      	    THEN METIS_TAC [normal_real])
      THEN METIS_TAC [extreal_le_eq, extreal_of_num_def])


THEN KNOW_TAC (``X x <> NegInf ``)
THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
THEN DISCH_TAC
THEN (`Y x <> NegInf ` by METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
THEN DISCH_TAC
THEN RW_TAC std_ss[]
THEN1 (METIS_TAC [extreal_real_lt])
THEN UNDISCH_TAC(``real (Y x) <= t``)
THEN REWRITE_TAC [real_def]
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
THEN1 (Q.EXISTS_TAC`real (Y x)`
      THEN METIS_TAC[normal_real])
THEN METIS_TAC [extreal_le_eq]);

val normal_real_mul = store_thm("normal_real_mul",
  `` !x y. x <> PosInf /\ x <> NegInf /\ y <> PosInf /\ y<> NegInf==>
           (Normal (real x * real y ) = Normal (real x) * Normal (real y))``,
REPEAT Cases
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss [real_def]
THEN METIS_TAC[extreal_mul_def]);


val rv_gt0_ninfinity_def = Define
`(rv_gt0_ninfinity [] = T) /\ (rv_gt0_ninfinity (h::t) = (!s. 0 <= h s /\ h s <> PosInf) /\ (rv_gt0_ninfinity t))   `;


val PROB_AND = store_thm("PROB_AND",
  ``  !p t X Y.
       rv_gt0_ninfinity [X; Y] /\
       indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s)) ==>
       (prob p (DFT_event p (D_AND X Y) t) =
        CDF p (\s. real (X s)) t * CDF p (\s. real (Y s)) t)``,

RW_TAC std_ss [AND_INTER, indep_var, DFT_event_def, rv_gt0_ninfinity_def]
THEN KNOW_TAC(``FINITE {0;1:num }``) THEN1(
     SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF])
THEN DISCH_TAC
THEN KNOW_TAC(``(prob p
        (BIGINTER {PREIMAGE ( if i = 0 then (\s. real (X s)) else
	              (\s. real (Y s))) ({u|u<=t:real}) INTER p_space p | i IN {0:num;1}}) =
         Normal
           (product {0:num;1}
             (\i. real (prob p (PREIMAGE ( if i = 0 then (\s. real (X s)) else
	                            (\s. real (Y s))) ( {u|u<=t:real})INTER p_space p)))))``)
      THEN1(
	   DEP_REWRITE_TAC [indep_varsD] 
	   THEN Q.EXISTS_TAC `\i.lborel`
	   THEN Q.EXISTS_TAC `{0;1:num}`
	   THEN RW_TAC std_ss[]
	   THEN1(
		FULL_SIMP_TAC std_ss [indep_vars, indep_sets])
	   THEN1(
		KNOW_TAC(``{0;1:num} = {x|(x=0:num)\/(x=1)}``) THEN1(
		        SRW_TAC [] [GSPECIFICATION, EXTENSION])
		THEN DISCH_TAC
		THEN ASM_REWRITE_TAC[])
	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
	   THEN RW_TAC std_ss [lborel_le])
THEN DISCH_TAC
THEN KNOW_TAC(`` {PREIMAGE
              (if i = 0:num then (\s. real (X s)) else (\s. real (Y s)))
              {u | u <= t} INTER p_space p |
               i IN {0; 1:num}} =  {PREIMAGE
               (\s. real ((X:'a->extreal) s)) {u|u<=t:real}  INTER p_space p;
	       PREIMAGE  (\s. real ((Y:'a->extreal) s)){u | u <= t} INTER p_space p }``)
	      THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION]
 	            THEN EQ_TAC
		    THEN1( SRW_TAC [] [GSPECIFICATION, EXTENSION]
		    	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
			   THEN SRW_TAC [] [GSPECIFICATION, EXTENSION])
	           THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
			THEN1(Q.EXISTS_TAC `0:num`
			      THEN RW_TAC std_ss[])
			THEN Q.EXISTS_TAC `1:num`
			THEN RW_TAC std_ss[])
THEN DISCH_TAC
THEN FULL_SIMP_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [BIGINTER_INTER]
THEN KNOW_TAC(``FINITE {0:num} /\ FINITE {1:num} /\ DISJOINT {0:num} {1:num} /\ ({0:num;1} =
                 {0} UNION {1})``)
     THEN1( SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF, DISJOINT_DEF])
THEN DISCH_TAC
THEN FULL_SIMP_TAC  std_ss [PRODUCT_UNION]
THEN FULL_SIMP_TAC std_ss  [PRODUCT_SING]
THEN KNOW_TAC(``random_variable (\s. real ((X:'a->extreal) s)) p borel /\
                 random_variable  (\s. real (Y s)) p borel``)
     THEN1(FULL_SIMP_TAC std_ss [indep_vars] 
     	   THEN UNDISCH_TAC(`` !i.
        i:num IN {x | (x = 0) \/ (x = 1)} ==>
        random_variable
          (if i = 0 then (\s. real (X s)) else (\s. real (Y s))) p
          (m_space lborel,measurable_sets lborel)``)
	  THEN DISCH_TAC
	  THEN RW_TAC std_ss []
	  THEN1(POP_ASSUM (MP_TAC o Q.SPEC `0:num `)
	  	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
	THEN POP_ASSUM (MP_TAC o Q.SPEC `1:num `)
	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
THEN DISCH_TAC
THEN KNOW_TAC(``(PREIMAGE (\s. real ((Y:'a->extreal) s)) {u | u <= t:real} INTER p_space p)
                  IN events p``)
     THEN1( REWRITE_TAC [events_def]
               THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``) THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
		THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
		THEN REWRITE_TAC [p_space_def, m_space_def]
		THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_le,lborel, borel, sigma_def])
THEN KNOW_TAC(``(PREIMAGE (\s. real ((X:'a->extreal) s)) {u | u <= t:real} INTER p_space p)
                  IN events p``)
      THEN1( REWRITE_TAC [events_def]
               THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``)
	            THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
		THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
		THEN REWRITE_TAC [p_space_def, m_space_def]
		THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_le,lborel, borel, sigma_def])
THEN DISCH_TAC
THEN DISCH_TAC

THEN KNOW_TAC(``prob_space p``) THEN1(FULL_SIMP_TAC std_ss [indep_vars, indep_sets])
THEN DISCH_TAC
 THEN FULL_SIMP_TAC std_ss [normal_real_mul, PROB_FINITE]
THEN FULL_SIMP_TAC std_ss [normal_real, PROB_FINITE]
THEN KNOW_TAC(``{s | (X:'a->extreal) s <= Normal t} = PREIMAGE X {s|s<= Normal t}``) THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN KNOW_TAC(``{s | (Y:'a->extreal) s <= Normal t} = PREIMAGE Y {s|s<= Normal t}``) THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL]
THEN RW_TAC std_ss [CDF_def, distribution_def]);

val PROB_AND = store_thm("PROB_AND",
  ``  !p t X Y.
       rv_gt0_ninfinity [X; Y] /\
       indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s)) ==>
       (prob p (DFT_event p (D_AND X Y) t) =
        CDF p (\s. real (X s)) t * CDF p (\s. real (Y s)) t)``,

RW_TAC std_ss [AND_INTER, indep_var, DFT_event_def, rv_gt0_ninfinity_def]
THEN KNOW_TAC(``FINITE {0;1:num }``) THEN1(
     SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF])
THEN DISCH_TAC
THEN KNOW_TAC(``(prob p
        (BIGINTER {PREIMAGE ( if i = 0 then (\s. real (X s)) else
	              (\s. real (Y s))) ({u|u<=t:real}) INTER p_space p | i IN {0:num;1}}) =
         Normal
           (product {0:num;1}
             (\i. real (prob p (PREIMAGE ( if i = 0 then (\s. real (X s)) else
	                            (\s. real (Y s))) ( {u|u<=t:real})INTER p_space p)))))``)
      THEN1(
	   DEP_REWRITE_TAC [indep_varsD] 
	   THEN Q.EXISTS_TAC `\i.lborel`
	   THEN Q.EXISTS_TAC `{0;1:num}`
	   THEN RW_TAC std_ss[]
	   THEN1(
		FULL_SIMP_TAC std_ss [indep_vars, indep_sets])
	   THEN1(
		KNOW_TAC(``{0;1:num} = {x|(x=0:num)\/(x=1)}``) THEN1(
		        SRW_TAC [] [GSPECIFICATION, EXTENSION])
		THEN DISCH_TAC
		THEN ASM_REWRITE_TAC[])
	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
	   THEN RW_TAC std_ss [lborel_le])
THEN DISCH_TAC
THEN KNOW_TAC(`` {PREIMAGE
              (if i = 0:num then (\s. real (X s)) else (\s. real (Y s)))
              {u | u <= t} INTER p_space p |
               i IN {0; 1:num}} =  {PREIMAGE
               (\s. real ((X:'a->extreal) s)) {u|u<=t:real}  INTER p_space p;
	       PREIMAGE  (\s. real ((Y:'a->extreal) s)){u | u <= t} INTER p_space p }``)
	      THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION]
 	            THEN EQ_TAC
		    THEN1( SRW_TAC [] [GSPECIFICATION, EXTENSION]
		    	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
			   THEN SRW_TAC [] [GSPECIFICATION, EXTENSION])
	           THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
			THEN1(Q.EXISTS_TAC `0:num`
			      THEN RW_TAC std_ss[])
			THEN Q.EXISTS_TAC `1:num`
			THEN RW_TAC std_ss[])
THEN DISCH_TAC
THEN FULL_SIMP_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [BIGINTER_INTER]
THEN KNOW_TAC(``FINITE {0:num} /\ FINITE {1:num} /\ DISJOINT {0:num} {1:num} /\ ({0:num;1} =
                 {0} UNION {1})``)
     THEN1( SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF, DISJOINT_DEF])
THEN DISCH_TAC
THEN FULL_SIMP_TAC  std_ss [PRODUCT_UNION]
THEN FULL_SIMP_TAC std_ss  [PRODUCT_SING]
THEN KNOW_TAC(``random_variable (\s. real ((X:'a->extreal) s)) p borel /\
                 random_variable  (\s. real (Y s)) p borel``)
     THEN1(FULL_SIMP_TAC std_ss [indep_vars] 
     	   THEN UNDISCH_TAC(`` !i.
        i:num IN {x | (x = 0) \/ (x = 1)} ==>
        random_variable
          (if i = 0 then (\s. real (X s)) else (\s. real (Y s))) p
          (m_space lborel,measurable_sets lborel)``)
	  THEN DISCH_TAC
	  THEN RW_TAC std_ss []
	  THEN1(POP_ASSUM (MP_TAC o Q.SPEC `0:num `)
	  	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
	THEN POP_ASSUM (MP_TAC o Q.SPEC `1:num `)
	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
THEN DISCH_TAC
THEN KNOW_TAC(``(PREIMAGE (\s. real ((Y:'a->extreal) s)) {u | u <= t:real} INTER p_space p)
                  IN events p``)
     THEN1( REWRITE_TAC [events_def]
               THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``) THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
		THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
		THEN REWRITE_TAC [p_space_def, m_space_def]
		THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_le,lborel, borel, sigma_def])
THEN KNOW_TAC(``(PREIMAGE (\s. real ((X:'a->extreal) s)) {u | u <= t:real} INTER p_space p)
                  IN events p``)
      THEN1( REWRITE_TAC [events_def]
               THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``)
	            THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
		THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
		THEN REWRITE_TAC [p_space_def, m_space_def]
		THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_le,lborel, borel, sigma_def])
THEN DISCH_TAC
THEN DISCH_TAC

THEN KNOW_TAC(``prob_space p``) THEN1(FULL_SIMP_TAC std_ss [indep_vars, indep_sets])
THEN DISCH_TAC
 THEN FULL_SIMP_TAC std_ss [normal_real_mul, PROB_FINITE]
THEN FULL_SIMP_TAC std_ss [normal_real, PROB_FINITE]
THEN KNOW_TAC(``{s | (X:'a->extreal) s <= Normal t} = PREIMAGE X {s|s<= Normal t}``) THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN KNOW_TAC(``{s | (Y:'a->extreal) s <= Normal t} = PREIMAGE Y {s|s<= Normal t}``) THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL]
THEN RW_TAC std_ss [CDF_def, distribution_def]);




val All_distinct_events_def = Define
`All_distinct_events p L t =  ALL_DISTINCT (MAP (\a. DFT_event p a t) L)`;


val CDF_prob_DFT_event = store_thm("CDF_prob_DFT_event",
  `` ! X p t.(!s. X s <> PosInf /\ 0<= X s)==>( CDF p (\s. real(X s)) t = prob p (DFT_event p X t))``,
RW_TAC std_ss [CDF_def, DFT_event_def, distribution_def]
THEN KNOW_TAC(``PREIMAGE (\s. real (X s)) {y | y <= t} INTER p_space p =
                  ({s | X s <= Normal t} INTER p_space p)``)
     THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]
           THEN RW_TAC std_ss [real_def]
	        THEN1( METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_of_num_def])
           THEN SELECT_ELIM_TAC
           THEN RW_TAC std_ss[] THEN1( Q.EXISTS_TAC `real (X x)`
                                        THEN METIS_TAC [normal_real])
           THEN RW_TAC std_ss[extreal_le_def])
THEN RW_TAC std_ss[]);


val PIE_lem2_or = store_thm("PIE_lem2_or",
  ``!f X Y . ALL_DISTINCT [X;Y] /\
       	     (!x.
         x IN {{X}; {Y}; {X; Y}} ==>
         f x <> PosInf) ==>
     (EXTREAL_SUM_IMAGE f {{X};{Y};{X;Y}} =		
      f {X} + f {Y} + f {X;Y})``,
RW_TAC std_ss[ALL_DISTINCT,MEM]
THEN MP_TAC(Q.ISPECL [`f:('a->bool)->extreal`,`{{Y}:'a->bool;  {X; Y}}`]
      EXTREAL_SUM_IMAGE_PROPERTY_POS)
THEN FULL_SIMP_TAC std_ss[prove(``!X Y . FINITE {{Y};  {X; Y}}``, SRW_TAC[][FINITE_SING])]
THEN DISCH_THEN (MP_TAC o Q.SPECL [`{X}:'a->bool`])
THEN FULL_SIMP_TAC std_ss[]
THEN RW_TAC std_ss[prove(``!X Y . X <> Y ==> (({{Y};  {X; Y}} DELETE {X}) = ({{Y}; {X; Y}}))``, SRW_TAC[][EXTENSION,GSPECIFICATION,ALL_DISTINCT,MEM] THEN METIS_TAC[])]
(*---------solving for {Y}---*)
THEN MP_TAC(Q.ISPECL [`f:('a->bool)->extreal`,`{ {X; Y}:'a->bool}`]
      EXTREAL_SUM_IMAGE_PROPERTY_POS)
THEN FULL_SIMP_TAC std_ss[prove(``!X Y. FINITE { {X; Y}}``, SRW_TAC[][FINITE_SING])]
THEN DISCH_THEN (MP_TAC o Q.SPECL [`{Y}:'a->bool`])
THEN FULL_SIMP_TAC std_ss[]
THEN RW_TAC std_ss[prove(``!X Y. X <> Y ==> (({ {X; Y}} DELETE {Y}) = ({ {X; Y}}))``, SRW_TAC[][EXTENSION,GSPECIFICATION,ALL_DISTINCT,MEM] THEN METIS_TAC[])]
THEN POP_ASSUM MP_TAC
THEN `(!x.
         x IN {{Y}; {X; Y}} ==>
         f x<> PosInf)` by 
(MATCH_MP_TAC (prove(``!X Y f. (!x.
             x IN {{X}; {Y};  {X:'a; Y}} ==>
             f x <>PosInf) ==> (!x.
         x IN {{Y};  {X; Y}} ==>
         f x <> PosInf)``, SRW_TAC[][EXTENSION,GSPECIFICATION]THEN METIS_TAC[]))
      THEN RW_TAC std_ss[])
THEN FULL_SIMP_TAC std_ss[]
THEN DISCH_TAC
THEN RW_TAC std_ss [add_assoc, add_not_infty, EXTREAL_SUM_IMAGE_SING]
THEN DEP_REWRITE_TAC[add_assoc,extreal_hvgTheory.add_not_infty]
THEN DISJ2_TAC
THEN UNDISCH_TAC (`` !x. x IN {{X}; {Y}; {X; Y}} ==> f x <> PosInf``)
THEN DISCH_TAC
THEN RW_TAC std_ss[]
     THEN1( POP_ASSUM (MP_TAC o Q.SPECL [`{X}:'a->bool`])
     	    THEN SRW_TAC [] [GSPECIFICATION, EXTENSION])
     THEN1(POP_ASSUM (MP_TAC o Q.SPECL [`{Y}:'a->bool`])
     	    THEN SRW_TAC [] [GSPECIFICATION, EXTENSION])
     THEN POP_ASSUM (MP_TAC o Q.SPECL [`{X;Y}:'a->bool`])
     	    THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]); 



val PROB_OR = store_thm ("PROB_OR",
  `` !X Y p t.
       rv_gt0_ninfinity [X; Y] /\ All_distinct_events  p [X; Y]  t /\
       indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s)) ==>
       (prob p (DFT_event p (D_OR X Y) t) =
        CDF p (\s. real (X s)) t + CDF p (\s. real (Y s)) t -
        CDF p (\s. real (X s)) t * CDF p (\s. real (Y s)) t)``,

RW_TAC std_ss[OR_UNION, All_distinct_events_def, rv_gt0_ninfinity_def]
THEN FULL_SIMP_TAC std_ss [MAP]
THEN KNOW_TAC(``random_variable (\s. real ((X:'a->extreal) s)) p borel /\
                 random_variable  (\s. real (Y s)) p borel``)
     THEN1(FULL_SIMP_TAC std_ss [indep_var, indep_vars] 
     	   THEN UNDISCH_TAC(`` !i.
        i:num IN {x | (x = 0) \/ (x = 1)} ==>
        random_variable
          (if i = 0 then (\s. real (X s)) else (\s. real (Y s))) p
          (m_space lborel,measurable_sets lborel)``)
	  THEN DISCH_TAC
	  THEN RW_TAC std_ss []
	  THEN1(POP_ASSUM (MP_TAC o Q.SPEC `0:num `)
	  	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
	THEN POP_ASSUM (MP_TAC o Q.SPEC `1:num `)
	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
THEN DISCH_TAC
THEN KNOW_TAC(``prob_space p``)
    THEN1(FULL_SIMP_TAC std_ss[indep_var, indep_vars, indep_sets])

THEN DISCH_TAC
THEN KNOW_TAC(``DFT_event p X t IN events p``)
     THEN1(RW_TAC std_ss [DFT_event_def]
      THEN KNOW_TAC(``{s:'a | X s <= Normal t}=PREIMAGE X {s |  s <= Normal t}``) THEN1(
      	             SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
      THEN DISCH_TAC
      THEN ASM_REWRITE_TAC[]
      THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL]
      THEN REWRITE_TAC [events_def]
      THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``) THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
      THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
      THEN REWRITE_TAC [p_space_def, m_space_def]
      THEN KNOW_TAC(``prob_space p``) THEN1(FULL_SIMP_TAC std_ss [indep_var, indep_vars, indep_sets])
      THEN DISCH_TAC
      THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_le,lborel, borel, sigma_def, indep_var, indep_vars, indep_sets])
THEN DISCH_TAC

THEN KNOW_TAC(``DFT_event p Y t IN events p``)
     THEN1(RW_TAC std_ss [DFT_event_def]
      THEN KNOW_TAC(``{s:'a | Y s <= Normal t}=PREIMAGE Y {s |  s <= Normal t}``) THEN1(
      	             SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
      THEN DISCH_TAC
      THEN ASM_REWRITE_TAC[]
      THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL]
      THEN REWRITE_TAC [events_def]
      THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``) THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
      THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
      THEN REWRITE_TAC [p_space_def, m_space_def]
      THEN KNOW_TAC(``prob_space p``) THEN1(FULL_SIMP_TAC std_ss [indep_var, indep_vars, indep_sets])
      THEN DISCH_TAC
      THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_le,lborel, borel, sigma_def, indep_var, indep_vars, indep_sets])
THEN DISCH_TAC

THEN KNOW_TAC(``DFT_event p X t UNION DFT_event p Y t =
                 union_list[DFT_event p X t; DFT_event p Y t]``)
     THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, union_list_def])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN DEP_REWRITE_TAC[PROB_INCLUSION_EXCLUSION_PRINCIPLE]
THEN RW_TAC list_ss[]
     THEN1( RW_TAC std_ss[])
     THEN1(RW_TAC std_ss[])
THEN RW_TAC std_ss [sum_set_def]
THEN KNOW_TAC(``{t' | t' SUBSET {DFT_event p X t; DFT_event p Y t} /\ t' <> {}}  =
                 {{DFT_event p X t};{DFT_event p Y t}; {DFT_event p X t; DFT_event p Y t}}``)
     THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION]
     THEN EQ_TAC
     	  THEN1(ASM_SET_TAC[])
     	  THEN ASM_SET_TAC[])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN DEP_REWRITE_TAC[PIE_lem2_or]
THEN RW_TAC std_ss[]
THEN1  (RW_TAC std_ss[extreal_pow_def]
       THEN DEP_REWRITE_TAC[PIE_lemma6]
       THEN DEP_REWRITE_TAC[PIE_lemma3]

THEN DEP_REWRITE_TAC[PROB_FINITE,EVENTS_COUNTABLE_INTER,COUNTABLE_IMAGE,finite_countable,IMAGE_FINITE]
       THEN (PSET_TAC[FINITE_SING,EVENTS_INTER,NOT_SING_EMPTY] THEN SRW_TAC[][FINITE_SING] THEN METIS_TAC[EVENTS_INTER]))
THEN RW_TAC std_ss[]
THEN RW_TAC real_ss[BIGINTER_SING,CARD_SING,CARD_INSERT,GSYM ADD1]
THEN FULL_SIMP_TAC list_ss[ALL_DISTINCT]
THEN PSET_TAC[]
THEN RW_TAC real_ss[extreal_pow_def]
THEN SRW_TAC[][FINITE_SING]
THEN RW_TAC real_ss[pow,prove (``!A B:extreal->bool. A <> B  ==> ((CARD {A;B} = SUC (SUC 0)) )``, RW_TAC list_ss[ALL_DISTINCT] THEN DEP_REWRITE_TAC[CARD_DEF] THEN SRW_TAC[][FINITE_SING])]
THEN DEP_REWRITE_TAC[PIE_lemma5]
THEN DEP_REWRITE_TAC[PROB_FINITE,EVENTS_COUNTABLE_INTER,COUNTABLE_IMAGE,finite_countable,IMAGE_FINITE]
THEN RW_TAC std_ss[] THEN  PSET_TAC[FINITE_SING,EVENTS_INTER,NOT_SING_EMPTY] THEN SRW_TAC[][FINITE_SING]
THEN RW_TAC std_ss[prove(``Normal 1 = 1``, RW_TAC std_ss[extreal_of_num_def] THEN METIS_TAC[]),mul_lone]
THEN RW_TAC std_ss[GSYM CDF_prob_DFT_event]
THEN RW_TAC std_ss [GSYM AND_INTER]
THEN DEP_REWRITE_TAC[PROB_AND]
THEN RW_TAC std_ss [rv_gt0_ninfinity_def]
THEN DEP_REWRITE_TAC[prove (``!a b c .  (b <> NegInf /\ b <> PosInf) /\ (c <> NegInf /\ c <> PosInf)  ==> (a + b + -c = a + b - c)``, REPEAT Cases THEN RW_TAC std_ss[extreal_of_num_def,extreal_add_def,extreal_sub_def,GSYM extreal_sub_add])]
THEN REWRITE_TAC [CDF_def, distribution_def]
THEN FULL_SIMP_TAC std_ss [DFT_event_def]
THEN KNOW_TAC(``({s|X s<=Normal t} = PREIMAGE (X:'a->extreal) {s|s<=Normal t}) /\ ({s|Y s<=Normal t} = PREIMAGE (Y:'a->extreal){s|s<=Normal t})``)
THEN1(SRW_TAC[] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN KNOW_TAC(``PREIMAGE (\s. real ((X:'a->extreal) s)) {y | y <= t} INTER p_space p IN events p /\                 PREIMAGE (\s. real ((Y:'a->extreal) s)) {y | y <= t} INTER p_space p IN events p``)
THEN1(METIS_TAC[PREIMAGE_EXTREAL_REAL])
THEN DISCH_TAC
THEN RW_TAC std_ss [PROB_FINITE]
THEN1( METIS_TAC [PROB_FINITE, mul_not_infty, extreal_of_num_def, extreal_mul_def, mul_extreal_not_infty])
THEN METIS_TAC [PROB_FINITE, mul_not_infty, extreal_of_num_def, extreal_mul_def, mul_extreal_not_infty]);


val DFT_event_AND_BEFORE = store_thm("DFT_event_AND_BEFORE",
  ``! p t X Y. DFT_event p (D_AND Y (D_BEFORE X Y)) t =
               {u|X(u) < Y(u) /\ Y(u) <= Normal t} INTER p_space p``,
RW_TAC std_ss [AND_INTER]
THEN RW_TAC std_ss [DFT_event_def, D_BEFORE_def]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty]);

val DFT_event_AND_BEFORE_PREIMAGE = store_thm("DFT_event_AND_BEFORE_PREIMAGE",
  `` ! p t X Y. DFT_event p (D_AND Y (D_BEFORE X Y)) t =
                PREIMAGE (\x. (X x, Y x)){(u,w)|u < w /\w <= Normal t} INTER p_space p``,

RW_TAC std_ss [DFT_event_AND_BEFORE, PREIMAGE_def]
THEN `(X x,Y x) IN {(u,w) | u <w /\ w<=Normal t} =
      ((X x,Y x) IN {(u,w) | (\a b. a < b /\ b <= Normal t )u w} )` by RW_TAC std_ss[]

THEN FULL_SIMP_TAC std_ss [IN_ELIM_PAIR_THM]);

val DFT_event_AND_BEFORE_PREIMAGE_GT0 = store_thm(
  "DFT_event_AND_BEFORE_PREIMAGE_GT0",
  ``   !p t X Y.
       (!s. 0 <= Y s) ==>
       (DFT_event p (D_AND Y (D_BEFORE X Y)) t =
        PREIMAGE (\x. (X x,Y x))
          {(u,w) | u < w /\ 0 <= w /\ w <= Normal t} INTER p_space p)``,

RW_TAC std_ss [DFT_event_AND_BEFORE_PREIMAGE, PREIMAGE_def]
THEN `(X x,Y x) IN {(u,w) | u <w /\ 0<=w/\w<=Normal t} =
      ((X x,Y x) IN {(u,w) | (\a b. a < b /\0<=b/\ b <= Normal t )u w} )` by RW_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [IN_ELIM_PAIR_THM]);


val DFT_event_AND_BEFORE_PREIMAGE_GT0_REAL = store_thm("DFT_event_AFTER_PREIMAGE_GT0",
  `` !p t X Y.rv_gt0_ninfinity [X; Y] /\ 0 <= t==>
     (DFT_event p (D_AND Y (D_BEFORE X Y)) t =
     PREIMAGE (\x. (real (X x),real (Y x)))
         {(u,w) | u < w /\ 0 <= w /\ w <= t} INTER p_space p) ``,

RW_TAC std_ss [DFT_event_AND_BEFORE_PREIMAGE_GT0,rv_gt0_ninfinity_def]
THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL_2RV]);


val PROB_DFT_AFTER = store_thm("PROB_DFT_AFTER",
  ``!X Y p fy t. rv_gt0_ninfinity [X; Y] /\ 0<=t/\
       prob_space p /\ indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s)) /\
       distributed p lborel (\s. real (Y s)) fy /\ (!y. 0 <= fy y) /\
      cont_CDF p (\s. real (X s))  /\  measurable_CDF p (\s. real (X s))   ==>
       (prob p (DFT_event p (D_AND Y (D_BEFORE X Y)) t) = 
           
        pos_fn_integral lborel
          (\y. fy y * (indicator_fn {u | 0<=u/\u <= t} y * CDF p (\s. real (X s)) y))) ``,

RW_TAC std_ss[DFT_event_AND_BEFORE_PREIMAGE_GT0, rv_gt0_ninfinity_def]
THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL_2RV]
THEN RW_TAC std_ss[prob_after]);


val BEFORE_event= store_thm("BEFORE_event_GT0",
  `` ! X Y p t.  (DFT_event p (D_BEFORE X Y ) t =
                  {s | (X s <= Normal t) /\ X s < Y s} INTER p_space p )``,
RW_TAC std_ss [DFT_event_def, D_BEFORE_def]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN RW_TAC std_ss[]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty]);

val BEFORE_event_GT0 = store_thm("BEFORE_event_GT0",
  `` ! X Y p t. (!s. 0<=X s)==>
       (DFT_event p (D_BEFORE X Y ) t =
        {s | (X s <= Normal t)/\ 0<=X s /\ X s < Y s} INTER p_space p )``,

RW_TAC std_ss [DFT_event_def, D_BEFORE_def]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN RW_TAC std_ss[]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty]);


val DFT_BEFORE_event_GT0 = store_thm("DFT_BEFORE_event_GT0",
  `` !X Y p t. (!s. 0<=X s)==>
      (DFT_event p (D_BEFORE X Y ) t =
        PREIMAGE (\s. (X s, Y s))
	   {(w,u) |  0<= w /\ w <= Normal t/\  w <  u} INTER p_space p )``,

RW_TAC std_ss[BEFORE_event_GT0]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]
THEN EQ_TAC
     THEN1(ASM_SET_TAC[])
THEN ASM_SET_TAC[]);


val PREIMAGE_EXTREAL_REAL_2RV_BEFORE = store_thm("PREIMAGE_EXTREAL_REAL_2RV_BEFORE",
  `` !X Y t p.
       (!s. X s <> PosInf /\ 0 <= X s /\ Y s <> PosInf /\ 0 <= Y s) /\
       0 <= t ==>
       (PREIMAGE (\s. (X s,Y s))
          {(w,u) | 0 <= w /\ w < u /\ w <= Normal t} INTER p_space p =
        PREIMAGE (\s. (real (X s),real (Y s)))
          {(w,u) | 0 <= w /\ w < u /\ w <= t} INTER p_space p)``, 

RW_TAC std_ss [PREIMAGE_def]
THEN KNOW_TAC(`` !s. (X s,Y s) IN {(w,u) | 0 <= w /\ w < u /\ w <= Normal t} =
                      0<=X s /\ X s < Y s /\ X s <= Normal t``)
THEN1(RW_TAC std_ss [IN_ELIM_PAIR_THM])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN KNOW_TAC(`` !s. (real (X s),real(Y s)) IN {(w,u) | 0 <= w /\ w < u /\ w <= t} =
                      0<=real(X s) /\  real (X s) < real(Y s) /\  real(X s) <= t``)
THEN1(RW_TAC std_ss [IN_ELIM_PAIR_THM])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN EQ_TAC
THEN1 (RW_TAC std_ss[]
      THEN1 (KNOW_TAC (``X x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
      	    THEN DISCH_TAC
      	    THEN KNOW_TAC (``Y x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
      	    THEN DISCH_TAC
	    THEN KNOW_TAC(``0<=X x ``) THEN1(RW_TAC std_ss[])
      	    THEN DISCH_TAC
	    THEN KNOW_TAC(``0<> PosInf /\ 0 <> NegInf ``)THEN1(
	    METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
	    THEN DISCH_TAC
	    THEN KNOW_TAC(``X x <> PosInf``)THEN1(RW_TAC std_ss[])
	    THEN DISCH_TAC
	    THEN KNOW_TAC(``0 = real 0``)
	    	 THEN1(RW_TAC std_ss[real_def]	    
		       THEN SELECT_ELIM_TAC
		       THEN RW_TAC std_ss[]
		       	    THEN1(Q.EXISTS_TAC`0`
				  THEN METIS_TAC[normal_real, extreal_of_num_def])
			    THEN METIS_TAC[normal_real, extreal_of_num_def, extreal_eq_zero])
	         THEN DISCH_TAC
		 THEN ASM_REWRITE_TAC[]
	         THEN  METIS_TAC [extreal_real_le])
	THEN1(KNOW_TAC (``X x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
      	    THEN DISCH_TAC
      	    THEN KNOW_TAC (``Y x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
	    THEN DISCH_TAC
THEN METIS_TAC[extreal_real_lt])
THEN RW_TAC std_ss[real_def]
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss []
      THEN1 (Q.EXISTS_TAC `real (X x)`
      	    THEN METIS_TAC [normal_real])
THEN METIS_TAC [extreal_le_eq, extreal_of_num_def])
THEN DISCH_TAC
THEN RW_TAC std_ss[]
THEN1(KNOW_TAC (``X x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
      	    THEN DISCH_TAC
      	    THEN KNOW_TAC (``Y x <> NegInf ``)
      	    THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])

THEN DISCH_TAC
THEN METIS_TAC[extreal_real_lt])

THEN KNOW_TAC (``X x <> NegInf ``)
THEN1 (METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
THEN DISCH_TAC
THEN  (`Y x <> NegInf ` by METIS_TAC [extreal_real_le, lt_infty,extreal_of_num_def,extreal_le_def, extreal_lt_def, lt_le, le_lt, le_infty ])
THEN METIS_TAC [extreal_le_eq, extreal_of_num_def, normal_real]);


val BEFORE_PREIMAGE_event_GT0 = store_thm("BEFORE_PREIMAGE_event_GT0",
  ``   !X Y p t.
  rv_gt0_ninfinity [X; Y] /\ 0 <=t ==>
     (DFT_event p (D_BEFORE X Y) t =
      PREIMAGE (\s. (real (X s),real (Y s)))
        {(w,u) | 0 <= w /\ w < u /\ w <= t} INTER p_space p)``,

RW_TAC std_ss [DFT_BEFORE_event_GT0, rv_gt0_ninfinity_def,  PREIMAGE_EXTREAL_REAL_2RV_BEFORE]
THEN KNOW_TAC(`` ! w u t.  0 <= w /\ w <= Normal t /\ w < u =
                            0 <= w /\ w < u /\ w <= Normal t``)
 THEN1(METIS_TAC[CONJ_ASSOC, CONJ_COMM])
 THEN DISCH_TAC
 THEN ASM_REWRITE_TAC[]
 THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL_2RV_BEFORE]);


val PROB_DFT_BEFORE = store_thm("PROB_DFT_BEFORE",
  ``!X Y p fx t. rv_gt0_ninfinity [X; Y] /\ 0<=t/\ prob_space p /\
     indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s))  /\
     distributed p lborel (\s. real (X s)) fx /\
      (!x. 0 <= fx x) /\ measurable_CDF p (\s. real (Y s)) ==>
      (prob p
       (DFT_event p (D_BEFORE X Y ) t) =
      pos_fn_integral lborel
        (\x.  (fx x * (indicator_fn {w |0<=w/\ w <= t} x * (1-CDF p(\s .real ( Y s))x))))) ``,

RW_TAC std_ss [DFT_BEFORE_event_GT0, rv_gt0_ninfinity_def,  PREIMAGE_EXTREAL_REAL_2RV_BEFORE]
THEN  KNOW_TAC(`` ! w u t.  0 <= w /\ w <= Normal t /\ w < u = 0 <= w /\ w < u /\ w <= Normal t``)
 THEN1(METIS_TAC[CONJ_ASSOC, CONJ_COMM])
 THEN DISCH_TAC
 THEN ASM_REWRITE_TAC[]
 THEN RW_TAC std_ss [PREIMAGE_EXTREAL_REAL_2RV_BEFORE]
 THEN KNOW_TAC(``! w u t:real. 0 <= w /\ w < u /\ w <= t =  0 <= w /\ w <= t /\ w < u ``)
       THEN1(METIS_TAC [CONJ_COMM, CONJ_ASSOC])
  THEN DISCH_TAC
  THEN ASM_REWRITE_TAC[]
  THEN DEP_REWRITE_TAC [prob_before]
  THEN RW_TAC std_ss[]);



(* -------------------------------------------------------------------------- *)
(*                                                                            *)
(*                      Simplification Theorems                               *)
(*                                                                            *)
(* -------------------------------------------------------------------------- *)


val D_OR_ID = store_thm("D_OR_ID",
  `` !  X . D_OR X X    = X``,
RW_TAC std_ss [D_OR_def] THEN
RW_TAC std_ss [min_refl] THEN
RW_TAC std_ss [ETA_THM]);


val D_OR_COMM = store_thm("D_OR_COMM",
  `` ! X Y . D_OR X Y    = D_OR Y X ``,
REPEAT GEN_TAC
THEN REWRITE_TAC [D_OR_def]
THEN METIS_TAC[min_comm]);

val D_OR_ASSOC = store_thm("D_OR_ASSOC",
  `` ! X Y Z . D_OR ( D_OR X Y  ) Z   =   D_OR X (D_OR Y Z)  ``,
  REPEAT GEN_TAC
THEN REWRITE_TAC [D_OR_def]

THEN RW_TAC std_ss [extreal_min_def]
THEN KNOW_TAC (``!s. (if
     (if (X :'a -> extreal) s <= (Y :'a -> extreal) s then X s
      else Y s) <= (Z :'a -> extreal) s
   then
     if X s <= Y s then X s else Y s
   else Z s) = (if X s <= if Y s <= Z s then Y s else Z s then X s
   else if Y s <= Z s then Y s
   else Z s)``) THEN1( RW_TAC std_ss[]
                        THEN REPEAT(POP_ASSUM MP_TAC)
			THEN RW_TAC std_ss []
			THEN FULL_SIMP_TAC std_ss[]
			THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);

val D_AND_ID = store_thm("D_AND_ID",
  `` ! X . D_AND X X  =X``,

RW_TAC std_ss [D_AND_def]
THEN RW_TAC std_ss[max_refl]
THEN RW_TAC std_ss [ETA_THM]);

val D_OR_NEVER = store_thm("D_OR_NEVER",
  `` ! X . D_OR X NEVER =  X ``,
RW_TAC std_ss [D_OR_def, NEVER_def, min_infty]
THEN RW_TAC std_ss [ETA_THM]);

val D_AND_ALWAYS = store_thm("D_AND_ALWAYS",
  ``! X . (!s. 0 <=X s)==> ( D_AND X ALWAYS = X)``,
REPEAT GEN_TAC
THEN REWRITE_TAC [D_AND_def, ALWAYS_def]
THEN RW_TAC std_ss [extreal_max_def]
THEN KNOW_TAC (``!s.  ((if (X :'a -> extreal) s <= (0 :extreal) then (0 :extreal) else X s)=X s)``) THEN1(RW_TAC std_ss[]
        THEN METIS_TAC[lt_le, lte_trans, let_trans])

THEN RW_TAC std_ss [ETA_THM]);

val D_AND_COMM = store_thm("D_AND_COMM",
  ``!  X Y . D_AND X Y    = D_AND Y X ``,
RW_TAC std_ss [D_AND_def]
THEN RW_TAC std_ss [max_comm]);


val D_AND_ASSOC = store_thm("D_AND_ASSOC",
  `` !X Y Z .   (D_AND ( D_AND X Y ) Z  =   D_AND X (D_AND Y Z ) ) ``,

RW_TAC std_ss [D_AND_def]
THEN RW_TAC std_ss [extreal_max_def]
THEN KNOW_TAC (``!s. ( if
     (if (X :'a -> extreal) s <= (Y :'a -> extreal) s then Y s
      else X s) <= (Z :'a -> extreal) s
   then
     Z s
   else if X s <= Y s then Y s
   else X s) = (if X s <= if Y s <= Z s then Z s else Y s then
     if Y s <= Z s then Z s else Y s
   else X s)  ``) THEN1( RW_TAC std_ss[]
                          THEN REPEAT(POP_ASSUM MP_TAC)
			  THEN RW_TAC std_ss []
			  THEN FULL_SIMP_TAC std_ss[]
			  THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])

THEN RW_TAC std_ss[]);

val D_AND_NEVER = store_thm("D_AND_NEVER",
  `` ! X . D_AND X NEVER = NEVER``,

RW_TAC std_ss [D_AND_def, max_infty, NEVER_def]);


val  D_AND_OVER_OR = store_thm("D_AND_OVER_OR",
  ``! X Y Z . D_AND X (D_OR Y Z ) = D_OR (D_AND X Y ) ( D_AND X Z)``,
RW_TAC std_ss [D_AND_def, D_OR_def]

THEN KNOW_TAC (``!s. max ((X :'a -> extreal) s)
     (min ((Y :'a -> extreal) s) ((Z :'a -> extreal) s)) =min (max (X s) (Y s)) (max (X s) (Z s))``) THEN1(RW_TAC std_ss[extreal_min_def, extreal_max_def]
        THEN REPEAT(POP_ASSUM MP_TAC)
	THEN RW_TAC std_ss []
	THEN FULL_SIMP_TAC std_ss[]
	THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss []);

val D_OR_OVER_AND = store_thm("D_OR_OVER_AND",
  ``! X Y Z . D_OR X (D_AND Y Z ) = D_AND (D_OR X Y ) ( D_OR X Z)``,
RW_TAC std_ss [D_AND_def, D_OR_def]
THEN KNOW_TAC (``!s. min ((X :'a -> extreal) s)
     (max ((Y :'a -> extreal) s) ((Z :'a -> extreal) s)) =max (min (X s) (Y s)) (min (X s) (Z s)) ``) THEN1 (RW_TAC std_ss[extreal_min_def, extreal_max_def]
          THEN REPEAT(POP_ASSUM MP_TAC)
	  THEN RW_TAC std_ss []
	  THEN FULL_SIMP_TAC std_ss[]
	  THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss []);

val D_OR_ALWAYS = store_thm("D_OR_ALWAYS",
  `` ! X. (!s. 0 <= X s)==> (D_OR X ALWAYS = ALWAYS)``,
RW_TAC std_ss [D_OR_def, ALWAYS_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val D_ABSORB_AND = store_thm("D_ABSORB_AND",
  ``  ! X Y . D_OR X ( D_AND X Y ) = X``,
RW_TAC std_ss [D_OR_def, D_AND_def]
THEN RW_TAC std_ss [extreal_min_def, extreal_max_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val D_ABSORB_AND = store_thm("D_ABSORB_AND",
  ``! X Y . D_AND X (D_OR X Y ) =X``,
RW_TAC std_ss [D_OR_def, D_AND_def]
THEN RW_TAC std_ss [extreal_min_def, extreal_max_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val D_INCLUSIVE_BEFORE_def2 = store_thm("D_INCLUSIVE_BEFORE_def2",
  ``! X Y . D_INCLUSIVE_BEFORE X Y = D_OR (D_BEFORE X Y ) (D_SIMULT X Y)``,
REPEAT GEN_TAC
THEN REWRITE_TAC [D_OR_def, D_BEFORE_def, D_SIMULT_def, D_INCLUSIVE_BEFORE_def]
THEN KNOW_TAC(``!s. 
( if X s <= Y s then X s else PosInf) =
(
   min ((\s. if X s < Y s then X s else PosInf) s)
     ((\s. if X s = Y s then X s else PosInf) s))
``) THEN1( RW_TAC std_ss[extreal_min_def]
            THEN REPEAT(POP_ASSUM MP_TAC)
	    THEN RW_TAC std_ss []
	    THEN FULL_SIMP_TAC std_ss[]
	    THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);

val D_BEFORE_NON_COMM = store_thm("D_BEFORE_NON_COMM",
  `` ! X Y . D_AND (D_BEFORE X Y)(D_BEFORE Y X) = NEVER``,

REPEAT GEN_TAC
THEN REWRITE_TAC [D_AND_def, D_OR_def, D_BEFORE_def, NEVER_def]
THEN RW_TAC std_ss[]
THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val D_BEFORE_NON_ASSOC1 = store_thm("D_BEFORE_NON_ASSOC1",
  `` ! X Y Z . D_BEFORE X (D_BEFORE Y Z) = D_OR (D_BEFORE X Y)(D_AND (D_AND X Y ) ( D_OR (D_BEFORE Z Y) (D_SIMULT Z Y)) )``,
  
REPEAT GEN_TAC
THEN RW_TAC std_ss[D_AND_def ,D_BEFORE_def, D_OR_def, D_SIMULT_def]
THEN KNOW_TAC (``!s. (if X s < if Y s < Z s then Y s else PosInf then X s else PosInf) = (min (if X s < Y s then X s else PosInf)
     (max (max (X s) (Y s))
        (min (if Z s < Y s then Z s else PosInf)
           (if Z s = Y s then Y s else PosInf))))``)
	   THEN1( RW_TAC std_ss[extreal_min_def, extreal_max_def] 
	          THEN REPEAT(POP_ASSUM MP_TAC)
		  THEN RW_TAC std_ss []
		  THEN FULL_SIMP_TAC std_ss[]
		  THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);

val D_BEFORE_NON_ASSOC2 = store_thm("D_BEFORE_NON_ASSOC2",
  `` ! X Y Z . D_BEFORE X (D_BEFORE Y Z) = D_OR (D_BEFORE X Y)(D_AND (D_AND X Y ) ( D_INCLUSIVE_BEFORE Z Y  ))``,
REPEAT GEN_TAC
THEN REWRITE_TAC [D_INCLUSIVE_BEFORE_def2]
THEN REWRITE_TAC [D_BEFORE_NON_ASSOC1]);

val D_BEFORE_NON_ASSOC3 = store_thm("D_BEFORE_NON_ASSOC3",
  `` ! X Y Z . D_BEFORE (D_BEFORE X Y) Z = D_AND (D_BEFORE X Y)(D_BEFORE X Z)``,
REPEAT GEN_TAC
THEN REWRITE_TAC [D_BEFORE_def, D_AND_def]
THEN RW_TAC std_ss[]
THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


val D_LEFT_BEFORE_OVER_OR = store_thm("D_LEFT_BEFORE_OVER_OR",
  ``! X Y Z. D_BEFORE X (D_OR Y Z )  = D_AND (D_BEFORE X Y)(D_BEFORE X Z)``,
RW_TAC std_ss[D_BEFORE_def,D_AND_def , D_OR_def ]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val D_LEFT_BEFORE_OVER_AND = store_thm("D_LEFT_BEFORE_OVER_AND",
  `` ! X Y Z. D_BEFORE X (D_AND Y Z )  = D_OR (D_BEFORE X Y)(D_BEFORE X Z)``,
RW_TAC std_ss[D_BEFORE_def,D_AND_def , D_OR_def ]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.25*) 


val D_BEFORE_OVER_SIMULT = store_thm("D_BEFORE_OVER_SIMULT",
  `` ! X Y Z . D_BEFORE X (D_SIMULT Y Z ) = D_OR (D_OR (D_OR (D_AND X (D_BEFORE Y Z)) (D_AND X (D_BEFORE Z Y)) )(D_BEFORE X Y) )(D_BEFORE X Z)``,
RW_TAC std_ss[D_BEFORE_def,D_AND_def , D_OR_def, D_SIMULT_def ]
THEN KNOW_TAC (``!s. (if X s < if Y s = Z s then Z s else PosInf then X s else PosInf) = ( min
     (min
        (min (max (X s) (if Y s < Z s then Y s else PosInf))
           (max (X s) (if Z s < Y s then Z s else PosInf)))
        (if X s < Y s then X s else PosInf))
     (if X s < Z s then X s else PosInf))``) THEN1(  RW_TAC std_ss[]
                                                      THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);



(* 3.26*) 
val D_BEFORE_OVER_INCLUSIVE_BEFORE = store_thm("D_BEFORE_OVER_INCLUSIVE_BEFORE",
  ``  ! X Y Z . D_BEFORE X (D_INCLUSIVE_BEFORE Y Z)= D_OR (D_BEFORE X Y) (D_AND (D_AND X Y) (D_BEFORE Z Y) )``,  
RW_TAC std_ss[D_BEFORE_def,D_AND_def , D_OR_def, D_SIMULT_def , D_INCLUSIVE_BEFORE_def]
THEN KNOW_TAC (``!s.(if X s < if Y s <= Z s then Y s else PosInf then X s else PosInf) =  (min (if X s < Y s then X s else PosInf)
     (max (max (X s) (Y s)) (if Z s < Y s then Z s else PosInf)))``)
     THEN1( RW_TAC std_ss[]
            THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);

(*3.27*) 
val RIGHT_BEFORE_OVER_OR = store_thm("RIGHT_BEFORE_OVER_OR",
  ``  ! X Y Z. D_BEFORE (D_OR X Y) Z = D_OR (D_BEFORE X Z) (D_BEFORE Y Z)``,
RW_TAC std_ss[D_BEFORE_def,D_AND_def , D_OR_def]

THEN KNOW_TAC (``!s. (if min (X s) (Y s) < Z s then min (X s) (Y s) else PosInf)= ( min (if X s < Z s then X s else PosInf)
     (if Y s < Z s then Y s else PosInf))``) THEN1( RW_TAC std_ss[]
                                                     THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);


(* 3.28*) 

val RIGHT_BEFORE_OVER_AND = store_thm("RIGHT_BEFORE_OVER_AND",
  ``  ! X Y Z. D_BEFORE (D_AND X Y )Z = D_AND (D_BEFORE X Z)(D_BEFORE Y Z)``,

RW_TAC std_ss[D_BEFORE_def,D_AND_def ]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.29_a*)
val RIGHT_BEFORE_OVER_SIMULT1 = store_thm("RIGHT_BEFORE_OVER_SIMULT1",
  ``  ! X Y Z. D_BEFORE (D_SIMULT X Y)Z= D_AND (D_SIMULT X Y) (D_BEFORE X Z)``,
RW_TAC std_ss[D_BEFORE_def, D_SIMULT_def, D_AND_def]

THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.29_b*)

val RIGHT_BEFORE_OVER_SIMULT2 = store_thm("RIGHT_BEFORE_OVER_SIMULT2",
  `` ! X Y Z. D_BEFORE (D_SIMULT X Y)Z = D_AND (D_SIMULT X Y) (D_BEFORE Y Z)``,
RW_TAC std_ss[D_BEFORE_def, D_SIMULT_def, D_AND_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.29_c *)
val RIGHT_BEFORE_OVER_SIMULT3 = store_thm("RIGHT_BEFORE_OVER_SIMULT3",
  `` ! X Y Z. D_BEFORE (D_SIMULT X Y)Z = D_SIMULT (D_BEFORE X Z) (D_BEFORE Y Z)``,
RW_TAC std_ss[D_BEFORE_def, D_SIMULT_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.30*)

val D_RIGHT_BEFORE_OVER_INCLUSIVE_BEFORE = store_thm(
  "D_RIGHT_BEFORE_OVER_INCLUSIVE_BEFORE",
  ``  ! X Y Z. D_BEFORE (D_INCLUSIVE_BEFORE X Y ) Z = D_AND (D_INCLUSIVE_BEFORE X Y) (D_BEFORE X Z)``,

RW_TAC std_ss[D_BEFORE_def, D_AND_def, D_INCLUSIVE_BEFORE_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);



(* 3.31*) 

val LEFT_OR_OVER_BEFORE_ABSORB = store_thm("LEFT_OR_OVER_BEFORE_ABSORB",
  `` ! X Y. D_OR X (D_BEFORE X Y ) = X``,
RW_TAC std_ss[D_BEFORE_def, D_OR_def]

THEN METIS_TAC [ETA_THM, min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.32*)
val RIGHT_OR_OVER_BEFORE_ABSORB = store_thm("RIGHT_OR_OVER_BEFORE_ABSORB",
  `` ! X Y. D_OR (D_BEFORE X Y) Y = D_OR X Y``,

RW_TAC std_ss[D_BEFORE_def, D_OR_def]

THEN METIS_TAC [ETA_THM, min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.33*)

val LEFT_AND_OVER_BEFORE_ABSORB = store_thm("LEFT_AND_OVER_BEFORE_ABSORB",
  ``! X Y. D_AND X (D_BEFORE X Y) = D_BEFORE X Y``,
RW_TAC std_ss[D_BEFORE_def, D_AND_def]
THEN METIS_TAC [ETA_THM, min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.34*) 

val AND_BEFORE_ABSORB = store_thm("AND_BEFORE_ABSORB",
  ``  ! X Y Z. D_AND (D_AND (D_BEFORE X Y)(D_BEFORE Y Z))(D_BEFORE X Z) = D_AND (D_BEFORE X Y)(D_BEFORE Y Z)``,
RW_TAC std_ss [D_BEFORE_def, D_AND_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.35*)
val D_SIMULT_COMM = store_thm("D_SIMULT_COMM",
  ``! X Y. D_SIMULT X Y = D_SIMULT Y X``,
REPEAT GEN_TAC 
THEN REWRITE_TAC [D_SIMULT_def]
THEN METIS_TAC[]);

(* 3.37 *)
val D_SIMULT_NEVER = store_thm("D_SIMULT_NEVER",
  `` ! X . D_SIMULT X NEVER = NEVER``,
RW_TAC std_ss [D_SIMULT_def, NEVER_def]);


(* 3.38*) 
val D_SIMULT_ID = store_thm("D_SIMULT_ID",
  `` ! X  . D_SIMULT X X = X``,
RW_TAC std_ss [D_SIMULT_def, ETA_THM]);


(* 3.39_a*)
val D_LEFT_SImULT_OVER_OR = store_thm("D_LEFT_SImULT_OVER_OR",
  `` ! X Y Z. D_SIMULT X (D_OR Y Z) = D_OR (D_OR (D_AND (D_SIMULT X Y) (D_SIMULT Y Z) ) (D_AND (D_SIMULT X Y ) (D_BEFORE Y Z ) ) ) (D_AND (D_SIMULT X Z ) (D_BEFORE Z Y) ) ``,

RW_TAC std_ss [D_SIMULT_def, D_OR_def, D_AND_def, D_BEFORE_def]
THEN KNOW_TAC (``!s. (if X s = min (Y s) (Z s) then min (Y s) (Z s) else PosInf)= (min
     (min
        (max (if X s = Y s then Y s else PosInf)
           (if Y s = Z s then Z s else PosInf))
        (max (if X s = Y s then Y s else PosInf)
           (if Y s < Z s then Y s else PosInf)))
     (max (if X s = Z s then Z s else PosInf)
        (if Z s < Y s then Z s else PosInf)))``) THEN1( RW_TAC std_ss[]
	                                                 THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);

(*3.39_b*)

val D_LEFT_SIMULT_OVER_OR2 = store_thm("D_LEFT_SIMULT_OVER_OR2",
  `` ! X Y Z . D_SIMULT X (D_OR Y Z ) = D_OR (D_AND (D_SIMULT X Y) (D_INCLUSIVE_BEFORE Y Z) )(D_AND (D_SIMULT X Z)(D_INCLUSIVE_BEFORE Z Y))``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_OR_def, D_AND_def, D_SIMULT_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.40_a*)
val D_LEFT_SIMULT_OVER_AND1 = store_thm("D_LEFT_SIMULT_OVER_AND1",
  `` !X Y Z. D_SIMULT X (D_AND Y Z ) = D_OR(D_OR (D_AND (D_SIMULT X Y)(D_SIMULT Y Z))(D_AND (D_SIMULT X Y) (D_BEFORE Z Y) ))(D_AND (D_SIMULT X Z )(D_BEFORE Y Z))``,
RW_TAC std_ss [D_BEFORE_def, D_OR_def, D_AND_def, D_SIMULT_def]

THEN KNOW_TAC (``!s. (if X s = max (Y s) (Z s) then max (Y s) (Z s) else PosInf) = (min
     (min
        (max (if X s = Y s then Y s else PosInf)
           (if Y s = Z s then Z s else PosInf))
        (max (if X s = Y s then Y s else PosInf)
           (if Z s < Y s then Z s else PosInf)))
     (max (if X s = Z s then Z s else PosInf)
        (if Y s < Z s then Y s else PosInf)))``) THEN1(RW_TAC std_ss[]
	                                                 THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);

(* 3.40_b*)


val D_LEFT_SIMULT_OVER_AND2 = store_thm("D_LEFT_SIMULT_OVER_AND2",
  `` ! X Y Z. D_SIMULT X (D_AND Y Z) = D_OR (D_AND (D_SIMULT X Y)(D_INCLUSIVE_BEFORE Z Y))(D_AND (D_SIMULT X Z)(D_INCLUSIVE_BEFORE Y Z))``,
RW_TAC std_ss [D_SIMULT_def, D_BEFORE_def, D_INCLUSIVE_BEFORE_def, D_AND_def, D_OR_def]

THEN KNOW_TAC (``!s. (if X s = max (Y s) (Z s) then max (Y s) (Z s) else PosInf) =(min
     (max (if X s = Y s then Y s else PosInf)
        (if Z s <= Y s then Z s else PosInf))
     (max (if X s = Z s then Z s else PosInf)
        (if Y s <= Z s then Y s else PosInf)))``) THEN1( RW_TAC std_ss[]
	                                                  THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])

THEN RW_TAC std_ss[]);

(* 3 .41*) 

val D_LEFT_SIMULT_OVER_BEFORE = store_thm("D_LEFT_SIMULT_OVER_BEFORE",
  `` ! X Y Z . D_SIMULT X (D_BEFORE Y Z) = D_AND (D_SIMULT X Y) (D_BEFORE Y Z)``,
RW_TAC std_ss [D_SIMULT_def, D_BEFORE_def, D_AND_def]
THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.42*) 

val D_LEFT_SIMULT_OVER_INCLUSIVE_BEFORE = store_thm(
  "D_LEFT_SIMULT_OVER_INCLUSIVE_BEFORE",
  ``  ! X Y Z . D_SIMULT X (D_INCLUSIVE_BEFORE Y Z) = D_AND (D_SIMULT X Y)(D_INCLUSIVE_BEFORE Y Z)``,
RW_TAC std_ss [D_SIMULT_def, D_INCLUSIVE_BEFORE_def, D_AND_def]
THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.43*)

val D_LEFT_OR_OVER_SIMULT = store_thm("D_LEFT_OR_OVER_SIMULT",
  `` ! X Y. D_OR X (D_SIMULT X Y) = X``,
RW_TAC std_ss [D_SIMULT_def, D_OR_def]
THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.44*)

val D_LEFT_AND_OVER_SIMULT = store_thm("D_LEFT_AND_OVER_SIMULT",
  ``  ! X Y. D_AND X (D_SIMULT X Y) = D_SIMULT X Y``,
RW_TAC std_ss [D_SIMULT_def, D_AND_def]
THEN  METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.45*) 

val D_AND_SIMULT_ABSORB = store_thm("D_AND_SIMULT_ABSORB",
  `` ! X Y Z. D_AND (D_AND (D_SIMULT X Y)(D_SIMULT Y Z))(D_SIMULT X Z)= D_AND (D_SIMULT X Y)(D_SIMULT Y Z)``,
RW_TAC std_ss [D_AND_def, D_SIMULT_def]

THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.46*)
val D_INCLUSIVE_AND = store_thm("D_INCLUSIVE_AND",
  ``! X Y. D_AND (D_INCLUSIVE_BEFORE X Y)(D_INCLUSIVE_BEFORE Y X) = D_SIMULT X Y``,
RW_TAC std_ss [D_AND_def, D_INCLUSIVE_BEFORE_def, D_SIMULT_def]

THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.47*) 

val D_LEFT_INCLUSIVE_OVER_INCLUSIVE1 = store_thm(
  "D_LEFT_INCLUSIVE_OVER_INCLUSIVE1",
  `` ! X Y Z. D_INCLUSIVE_BEFORE X (D_INCLUSIVE_BEFORE Y Z ) = D_OR(D_OR (D_BEFORE X Y)(D_AND (D_AND X Y)(D_BEFORE Z Y)))(D_AND (D_SIMULT X Y)(D_INCLUSIVE_BEFORE Y Z))``, 
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_AND_def, D_SIMULT_def, D_OR_def, D_BEFORE_def]

THEN KNOW_TAC (``!s. (if X s <= if Y s <= Z s then Y s else PosInf then X s else PosInf) =( min
     (min (if X s < Y s then X s else PosInf)
        (max (max (X s) (Y s)) (if Z s < Y s then Z s else PosInf)))
     (max (if X s = Y s then Y s else PosInf)
        (if Y s <= Z s then Y s else PosInf)))``) THEN1( RW_TAC std_ss[]
	                                                  THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);

(* 3.47_b*)
val D_RIGHT_INCLUSIVE_OVER_INCLUSIVE = store_thm(
  "D_RIGHT_INCLUSIVE_OVER_INCLUSIVE",
  ``  ! X Y Z. D_INCLUSIVE_BEFORE  (D_INCLUSIVE_BEFORE X Y) Z = D_AND (D_INCLUSIVE_BEFORE X Y ) ( D_INCLUSIVE_BEFORE X Z)``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_AND_def]

THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);



(* 3.49*)

val NEVER_INCLUSIVE_BEFORE_BEFORE = store_thm("NEVER_INCLUSIVE_BEFORE_BEFORE",
  `` ! X. D_INCLUSIVE_BEFORE NEVER X = NEVER``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, NEVER_def]);


(* 3.50*)
val INCLUSIVE_BEFORE_NEVER = store_thm("INCLUSIVE_BEFORE_NEVER",
  `` ! X. D_INCLUSIVE_BEFORE X NEVER = X``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, NEVER_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.51*)
val D_INCLUSIVE_ID = store_thm("D_INCLUSIVE_ID",
  `` ! X. D_INCLUSIVE_BEFORE X X = X``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def] 
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.52*)
val D_LEFT_INCLUSIVE_OVER_OR = store_thm("D_LEFT_INCLUSIVE_OVER_OR",
  `` ! X Y Z. D_INCLUSIVE_BEFORE X (D_OR Y Z) = D_AND (D_INCLUSIVE_BEFORE X Y)(D_INCLUSIVE_BEFORE X Z)``,

RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_AND_def, D_OR_def]

THEN KNOW_TAC (``!s. (if X s <= min (Y s) (Z s) then X s else PosInf) =( max (if X s <= Y s then X s else PosInf)
     (if X s <= Z s then X s else PosInf))``) THEN1( RW_TAC std_ss[]
                                                       THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);


(* 3.53*)
val D_LEFT_INCLUSIVE_OVER_AND = store_thm("D_LEFT_INCLUSIVE_OVER_AND",
  ``! X Y Z. D_INCLUSIVE_BEFORE X (D_AND Y Z) = D_OR (D_INCLUSIVE_BEFORE X Y )(D_INCLUSIVE_BEFORE X Z)``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_AND_def, D_OR_def]
THEN KNOW_TAC (``!s.  (if X s <= max (Y s) (Z s) then X s else PosInf) =( min (if X s <= Y s then X s else PosInf)
     (if X s <= Z s then X s else PosInf))
``) THEN1( RW_TAC std_ss[]
            THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);


(* 3.54*)
val D_LEFT_INCLUSIVE_OVER_BEFORE = store_thm("D_LEFT_INCLUSIVE_OVER_BEFORE",
  `` ! X Y Z. D_INCLUSIVE_BEFORE X (D_BEFORE Y Z) = D_OR (D_OR (D_BEFORE X Y)(D_AND (D_AND X Y)(D_INCLUSIVE_BEFORE Z Y)))(D_AND (D_SIMULT X Y)(D_BEFORE Y Z))``,

RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_SIMULT_def, D_OR_def, D_AND_def, D_BEFORE_def]

THEN KNOW_TAC (``!s. (if X s <= if Y s < Z s then Y s else PosInf then X s else PosInf) = (min
     (min (if X s < Y s then X s else PosInf)
        (max (max (X s) (Y s)) (if Z s <= Y s then Z s else PosInf)))
     (max (if X s = Y s then Y s else PosInf)
        (if Y s < Z s then Y s else PosInf)))``) THEN1( RW_TAC std_ss[]
	                                                 THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);


(*3.55_a*)

val D_LEFT_INCLUSIVE_OVER_SIMULT = store_thm("D_LEFT_INCLUSIVE_OVER_SIMULT",
  ``  ! X Y Z. D_INCLUSIVE_BEFORE X (D_SIMULT Y Z) = D_OR(D_OR(D_OR (D_OR (D_AND X (D_BEFORE Y Z))(D_AND X (D_BEFORE Z Y)))(D_BEFORE X Y))(D_BEFORE X Z))(D_AND (D_SIMULT X Y)(D_SIMULT Y Z))``,
REPEAT GEN_TAC
THEN RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_SIMULT_def, D_OR_def, D_AND_def, D_BEFORE_def]
THEN KNOW_TAC (``!s. (if X s <= if Y s = Z s then Z s else PosInf then X s else PosInf) =(min
     (min
        (min
           (min (max (X s) (if Y s < Z s then Y s else PosInf))
              (max (X s) (if Z s < Y s then Z s else PosInf)))
           (if X s < Y s then X s else PosInf))
        (if X s < Z s then X s else PosInf))
     (max (if X s = Y s then Y s else PosInf)
        (if Y s = Z s then Z s else PosInf)))``) THEN1( RW_TAC std_ss[]
	                                                THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);


(* 3.56*)
val D_RIGHT_INCLUSIVE_OVER_SIMULT1 = store_thm("D_RIGHT_INCLUSIVE_OVER_SIMULT1",
  `` !X Y Z. D_INCLUSIVE_BEFORE (D_OR X Y) Z = D_OR (D_INCLUSIVE_BEFORE X Z)(D_INCLUSIVE_BEFORE Y Z)``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_OR_def]

THEN KNOW_TAC (``!s. (if min (X s) (Y s) <= Z s then min (X s) (Y s) else PosInf) =(min (if X s <= Z s then X s else PosInf)
     (if Y s <= Z s then Y s else PosInf))``) THEN1( RW_TAC std_ss[]
                                                      THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);

(* 3.58_a*)
val D_RIGHT_INCLUSIVE_OVER_SIMULT1 = store_thm("D_RIGHT_INCLUSIVE_OVER_SIMULT1",
  ``  ! X Y Z. D_INCLUSIVE_BEFORE (D_SIMULT X Y) Z= D_AND (D_SIMULT X Y)(D_INCLUSIVE_BEFORE X Z)``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_SIMULT_def, D_AND_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.58_b*)
val D_RIGHT_INCLUSIVE_OVER_SIMULT2 = store_thm("D_RIGHT_INCLUSIVE_OVER_SIMULT2",
  ``  ! X Y Z. D_INCLUSIVE_BEFORE (D_SIMULT X Y) Z= D_AND (D_SIMULT X Y)(D_INCLUSIVE_BEFORE Y Z)``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_SIMULT_def, D_AND_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.58_C*)

val D_RIGHT_INCLUSIVE_OVER_SIMULT3 = store_thm("D_RIGHT_INCLUSIVE_OVER_SIMULT3",
  `` ! X Y Z. D_INCLUSIVE_BEFORE (D_SIMULT X Y) Z= D_SIMULT (D_INCLUSIVE_BEFORE X Z)(D_INCLUSIVE_BEFORE Y Z)``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_SIMULT_def]
THEN KNOW_TAC (``!s. (if (if X s = Y s then Y s else PosInf) <= Z s then
     if X s = Y s then Y s else PosInf
   else PosInf) =(if
     (if X s <= Z s then X s else PosInf) =
     if Y s <= Z s then Y s else PosInf
   then
     if Y s <= Z s then Y s else PosInf
   else PosInf)``) THEN1 (RW_TAC std_ss[]
                           THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);

(* 3.59*)
val D_RIGHT_INCLUSIVE_OVER_BEFORE = store_thm("D_RIGHT_INCLUSIVE_OVER_BEFORE",
  `` ! X Y Z. D_INCLUSIVE_BEFORE (D_BEFORE X Y ) Z = D_AND (D_BEFORE X Y)(D_INCLUSIVE_BEFORE X Z)``,
RW_TAC std_ss[D_INCLUSIVE_BEFORE_def,D_AND_def, D_BEFORE_def] 
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.60*)
val D_LEFT_OR_INCLUSIVE_ABSORB = store_thm("D_LEFT_OR_INCLUSIVE_ABSORB",
  `` ! X Y. D_OR X (D_INCLUSIVE_BEFORE X Y) = X``,
RW_TAC std_ss [D_OR_def, D_INCLUSIVE_BEFORE_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);



(*3.61*)
val D_LEFT_OR_OVER_INCLUSIVE = store_thm("D_LEFT_OR_OVER_INCLUSIVE",
  ``  ! X Y. D_OR Y (D_INCLUSIVE_BEFORE X Y) = D_OR X Y``,
RW_TAC std_ss [D_OR_def, D_INCLUSIVE_BEFORE_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.62*) 
val D_LEFT_AND_OVER_INCLUSIVE = store_thm("D_LEFT_AND_OVER_INCLUSIVE",
  `` !X Y . D_AND X (D_INCLUSIVE_BEFORE X Y ) = D_INCLUSIVE_BEFORE X Y``,
RW_TAC std_ss [D_AND_def, D_INCLUSIVE_BEFORE_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.63*)
val D_INCLUSIVE_OR_ABSORB = store_thm("D_INCLUSIVE_OR_ABSORB",
  ``  ! X Y. D_OR (D_INCLUSIVE_BEFORE X Y)(D_INCLUSIVE_BEFORE Y X ) =D_OR X Y``,
RW_TAC std_ss [D_OR_def, D_INCLUSIVE_BEFORE_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.64*)
val Lemma_64 = store_thm("Lemma_64",
  `` ! X Y. D_OR (D_AND X (D_INCLUSIVE_BEFORE Y X))(D_AND Y (D_INCLUSIVE_BEFORE X Y)) = D_AND X Y``, 
RW_TAC std_ss [ D_OR_def,D_INCLUSIVE_BEFORE_def, D_AND_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.65*)

val Lemma_65 = store_thm("Lemma_65",
  ``  ! X Y. D_OR (D_INCLUSIVE_BEFORE X Y)(D_AND X(D_INCLUSIVE_BEFORE Y X)) = X``,
RW_TAC std_ss [ D_OR_def,D_INCLUSIVE_BEFORE_def, D_AND_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);



(* 3.66*)
val Lemma_66 = store_thm("Lemma_66",
  ``  ! X Y Z. D_AND (D_AND (D_INCLUSIVE_BEFORE X Y)(D_INCLUSIVE_BEFORE Y Z))(D_INCLUSIVE_BEFORE X Z) = D_AND (D_INCLUSIVE_BEFORE X Y)(D_INCLUSIVE_BEFORE Y Z)``,

RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_OR_def, D_AND_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(*3.67*)
val Lemma_67 = store_thm("Lemma_67",
  `` ! X Y. D_OR (D_INCLUSIVE_BEFORE X Y)(D_BEFORE X Y) = D_INCLUSIVE_BEFORE X Y``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_OR_def, D_BEFORE_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.68*)
val Lemma_68 = store_thm("Lemma_68",
  ``! X Y. D_OR (D_INCLUSIVE_BEFORE X Y)(D_SIMULT X Y) = D_INCLUSIVE_BEFORE X Y``,
RW_TAC std_ss [D_INCLUSIVE_BEFORE_def, D_OR_def, D_SIMULT_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.69*)

val Lemma_69 = store_thm("Lemma_69",
  ``! X Y. D_AND (D_BEFORE X Y)(D_SIMULT X Y) = NEVER``,
RW_TAC std_ss [D_BEFORE_def, D_AND_def, D_SIMULT_def, NEVER_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.70*)


val Lemma_70 = store_thm("Lemma_70",
  `` ! X Y Z. D_AND (D_BEFORE X Y)(D_SIMULT Y Z)= D_AND (D_BEFORE X Z)(D_SIMULT Y Z)``,
RW_TAC std_ss [D_BEFORE_def, D_AND_def, D_SIMULT_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.71*)

val Lemma_71 = store_thm("Lemma_71",
  `` ! X Y. D_AND (D_INCLUSIVE_BEFORE X Y)(D_BEFORE X Y) = D_BEFORE X Y``,
RW_TAC std_ss [D_BEFORE_def, D_AND_def, D_INCLUSIVE_BEFORE_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(* 3.72*)

val Lemma_72 = store_thm("Lemma_72",
  `` ! X Y. D_AND (D_BEFORE X Y)(D_INCLUSIVE_BEFORE Y X) = NEVER``,
  RW_TAC std_ss [D_BEFORE_def, D_AND_def, D_INCLUSIVE_BEFORE_def, NEVER_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(* 3.73*)

val Lemma_73 = store_thm("Lemma_73",
  ``  ! X Y. D_AND (D_INCLUSIVE_BEFORE X Y)(D_SIMULT X Y)  = D_SIMULT X Y``,
RW_TAC std_ss [D_SIMULT_def, D_AND_def, D_INCLUSIVE_BEFORE_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);



(*3.74*)
val Lemma_74 = store_thm("Lemma_74",
  `` ! X Y. D_OR(D_OR (D_BEFORE X Y)(D_SIMULT X Y))(D_BEFORE Y X)=D_OR X Y``,
RW_TAC std_ss [D_SIMULT_def, D_OR_def, D_BEFORE_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.75*) 
val Lemma_75 = store_thm("Lemma_75",
  `` ! X Y. D_OR(D_OR (D_AND X(D_BEFORE Y X))(D_SIMULT X Y))(D_AND Y (D_BEFORE X Y)) = D_AND X Y``,
RW_TAC std_ss [D_SIMULT_def, D_OR_def, D_BEFORE_def, D_AND_def]

THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

(*3.76*)

val Lemma_76 = store_thm("Lemma_76",
  `` ! X Y. D_OR(D_OR (D_BEFORE X Y)(D_SIMULT X Y))(D_AND X (D_BEFORE Y X)) = X``,
RW_TAC std_ss [D_SIMULT_def, D_OR_def, D_BEFORE_def, D_AND_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);
 
val Lemma_77 = store_thm("Lemma_77",
  `` ! X Y Z. D_AND (D_AND (D_BEFORE X Y)(D_BEFORE Y Z))(D_INCLUSIVE_BEFORE X Z) = D_AND (D_BEFORE X Y)(D_BEFORE Y Z)``,
RW_TAC std_ss [D_AND_def, D_BEFORE_def, D_INCLUSIVE_BEFORE_def]
THEN METIS_TAC  [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


(******************************************************************************)
(*                                                                            *)
(*   some required lemmas for  child theories                                 *)
(*                                                                            *)
(******************************************************************************)


val indep_vars_var = store_thm("indep_vars_var",
  ``! X  p M s t ii. s IN ii /\ t IN ii /\ s <> t ==>  (indep_vars p (\i. M) X ii ==>(indep_var p M (X s) M (X t)))  ``,
  

RW_TAC std_ss[]
THEN KNOW_TAC(``indep_vars p (\i. M) X {s;t:num} ``)
THEN1 ( `{s;t:num} SUBSET ii` by ASM_SET_TAC[]
       THEN METIS_TAC [indep_vars_subset])
THEN DISCH_TAC
THEN WEAKEN_TAC (equal(Term` indep_vars p (\i. M) X ii`))

THEN RW_TAC std_ss[indep_var]
THEN FULL_SIMP_TAC std_ss[indep_vars]
THEN RW_TAC std_ss[]
     THEN1(UNDISCH_TAC(``   !i:num.
             i IN {s; t} ==>
             random_variable (X i) p (m_space M,measurable_sets M)``)
	   THEN DISCH_TAC
	   THEN POP_ASSUM (MP_TAC o Q.SPEC `s:num`)
	   THEN ASM_SET_TAC[])
      THEN1(UNDISCH_TAC(``   !i:num.
             i IN {s; t} ==>
             random_variable (X i) p (m_space M,measurable_sets M)``)
	   THEN DISCH_TAC
	   THEN POP_ASSUM (MP_TAC o Q.SPEC `t:num`)
	   THEN ASM_SET_TAC[])
	   
THEN `{x:num| (x = 0) \/ (x = 1)}  = {0;1}` by ASM_SET_TAC[]

THEN ONCE_ASM_REWRITE_TAC[]

THEN RW_TAC std_ss [indep_sets]

       THEN1( FULL_SIMP_TAC std_ss[indep_sets])
       THEN1( FULL_SIMP_TAC std_ss [indep_sets]
       	      THEN  KNOW_TAC(``{PREIMAGE f (A:'b->bool) INTER p_space p |
   (f = X i) /\
   A IN measurable_sets M} = {PREIMAGE ((X:num->'a->'b) i) (A:'b->bool) INTER p_space p |
  A IN measurable_sets M}``)
                  THEN1( ASM_SET_TAC[])
	      THEN DISCH_TAC
	      THEN ASM_REWRITE_TAC[]
	      THEN ASM_SET_TAC[])
THEN FULL_SIMP_TAC std_ss[indep_sets]
THEN UNDISCH_TAC(``!J:num->bool.
             J SUBSET {s; t} /\ J <> {} /\ FINITE J ==>
             !A.
                 A IN
                 Pi J
                   (\i.
                        {PREIMAGE f A INTER p_space p |
                        (f = X i) /\ A IN measurable_sets M}) ==>
                 (prob p (BIGINTER {A j | j IN J}) =
                  Normal (product J (\j. real (prob p (A j)))))``)
		THEN DISCH_TAC
		THEN POP_ASSUM (MP_TAC o Q.SPEC `{s;t:num} `)
		THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN FULL_SIMP_TAC std_ss [Pi_iff]
		THEN KNOW_TAC(``!i.{PREIMAGE f A INTER p_space p |
             (f = (X:num->'a->'b) i) /\ A IN measurable_sets M} =
	     {PREIMAGE (X i) A INTER p_space p |
             A IN measurable_sets M} ``) THEN1(ASM_SET_TAC[])
 	        THEN DISCH_TAC
		THEN FULL_SIMP_TAC std_ss[]


		  THEN KNOW_TAC(``{PREIMAGE ((X:num->'a->'b) s) A INTER p_space p| A IN measurable_sets M } SUBSET events p``) THEN1(ASM_SET_TAC[])
		  THEN DISCH_TAC
		  THEN KNOW_TAC(``{PREIMAGE ((X:num->'a->'b) t) A INTER p_space p| A IN measurable_sets M } SUBSET events p``)THEN1(ASM_SET_TAC[])
		  THEN DISCH_TAC
		  THEN  `J SUBSET {0:num;1} = (J = {0:num} ) \/ ( J = {1:num}) \/ (J = {0:num;1})` by ASM_SET_TAC[]
		 THEN FULL_SIMP_TAC std_ss[]
(*3 subgoals*)		 THEN1( KNOW_TAC(``{(A:num->'a->bool) j | j IN {0}} = {A 0} ``)
			 	THEN1(SRW_TAC [] [EXTENSION, GSPECIFICATION])
				THEN DISCH_TAC
				THEN ASM_REWRITE_TAC[]
				THEN RW_TAC std_ss [PRODUCT_SING]
				THEN RW_TAC std_ss [BIGINTER_SING]
				THEN UNDISCH_TAC(`` !i:num.
             i IN {0} ==>
             (A:num->'a->bool) i IN
             {PREIMAGE f A INTER p_space p |
             (f = if i = 0 then (X:num->'a->'b) s else X t) /\ A IN measurable_sets M}``)
				THEN DISCH_TAC

				THEN POP_ASSUM (MP_TAC o Q.SPEC `0:num`)
				THEN `0:num IN {0:num} ` by ASM_SET_TAC[]
				THEN RW_TAC std_ss[]

				THEN `?A'.( A 0 = PREIMAGE (X s) A' INTER p_space p)/\ A' IN measurable_sets M` by 
				(  PSET_TAC[]
				 THEN ASM_REWRITE_TAC[])
				 THEN `random_variable (X s) p (m_space M, measurable_sets M)` by 
				 (UNDISCH_TAC(`` !i:num.
        i IN {s; t} ==>
       random_variable (X i) p
               (m_space M,measurable_sets M)``)

	   	      		      THEN DISCH_TAC
				      THEN POP_ASSUM (MP_TAC o Q.SPEC `s:num`)
				      THEN SRW_TAC[] [EXTENSION, GSPECIFICATION]
				      THEN FULL_SIMP_TAC std_ss[])
				 THEN `PREIMAGE (X s) (A':'b->bool) INTER p_space p IN events p` by 
				      (METIS_TAC [IN_MEASURABLE, measurable_sets_def, subsets_def, m_space_def, events_def, space_def, random_variable_def, sigma_def, p_space_def, IN_SING, SUBSET_DEF, GSPECIFICATION, EXTENSION, IN_DEF])
				 THEN METIS_TAC [PROB_FINITE, normal_real])

(* 2 subgoals A 1*)      	THEN1 (KNOW_TAC(``{(A:num->'a->bool) j | j IN {1}} = {A 1} ``)
			 	THEN1(SRW_TAC [] [EXTENSION, GSPECIFICATION])
				THEN DISCH_TAC
				THEN ASM_REWRITE_TAC[]
				THEN RW_TAC std_ss [PRODUCT_SING]
				THEN RW_TAC std_ss [BIGINTER_SING]
				THEN UNDISCH_TAC(`` !i:num.
				        i IN {1} ==>
			        A i IN
			        {PREIMAGE f A INTER p_space p |
				         (f = if i = 0 then ((X:num->'a->'b) s) else (X t)) /\ A								 IN measurable_sets M}``)
				THEN DISCH_TAC

				THEN POP_ASSUM (MP_TAC o Q.SPEC `1:num`)
				THEN `1:num IN {1:num} ` by ASM_SET_TAC[]
				THEN RW_TAC std_ss[]

				THEN `?A'.( A 1 = PREIMAGE (X t) A' INTER p_space p)/\ A' IN measurable_sets M` by 
				( PSET_TAC[]
				 THEN ASM_REWRITE_TAC[])
				 THEN `random_variable (X t) p (m_space M, measurable_sets M)` by 
				 (UNDISCH_TAC(`` !i:num.
       i IN {s; t} ==>
             random_variable ((X:num->'a->'b) i) p
               (m_space M,measurable_sets M)``)

	   	      		      THEN DISCH_TAC
				      THEN POP_ASSUM (MP_TAC o Q.SPEC `t:num`)
				      THEN SRW_TAC[] [EXTENSION, GSPECIFICATION]
				      THEN FULL_SIMP_TAC std_ss[])
				 THEN `PREIMAGE (X t) (A':'b->bool) INTER p_space p IN events p` by   
				        (METIS_TAC [IN_MEASURABLE, measurable_sets_def, subsets_def, m_space_def, events_def, space_def, lborel, borel,             random_variable_def, sigma_def, p_space_def, IN_SING, SUBSET_DEF, GSPECIFICATION, EXTENSION, IN_DEF])
				 THEN METIS_TAC [PROB_FINITE, normal_real])


(* 1 subgoal {0;1}*)
THEN KNOW_TAC(``{(A:num->'a->bool) j | j IN {0:num; 1}}={A 0; A 1}``) THEN1(ASM_SET_TAC[])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN `FINITE {0:num} /\ FINITE {1:num} /\ DISJOINT {0:num} {1:num}/\ ({0;1:num} = {0} UNION {1})` by SRW_TAC  [] [GSPECIFICATION, EXTENSION, FINITE_DEF, DISJOINT_DEF]
THEN RW_TAC std_ss [BIGINTER_INTER,PRODUCT_UNION, PRODUCT_SING]

THEN KNOW_TAC(``?A'.( (A:num->'a->bool) 0 = PREIMAGE ((X:num->'a->'b) s) A' INTER p_space p)/\ A' IN measurable_sets M``) 
     THEN1 (UNDISCH_TAC(``!i.
             i IN {0; 1:num} ==>
             (A:num->'a->bool) i IN
             {PREIMAGE f A INTER p_space p |
             (f = if i = 0 then (X:num->'a->'b) s else X t) /\ A IN measurable_sets M}``)
     THEN DISCH_TAC
     THEN POP_ASSUM (MP_TAC o Q.SPEC `(0:num) `)
     THEN FULL_SIMP_TAC std_ss[prove(``0 IN {0;(1:num)}``,
     SRW_TAC[][EXTENSION,GSPECIFICATION])]
      THEN `A 0 IN
	 {PREIMAGE f A INTER p_space p |
	  (f = X s) /\ A IN measurable_sets M} = A 0 IN
	  {PREIMAGE (X s) A INTER p_space p|
	  A IN measurable_sets M }` by  ASM_SET_TAC[]
         THEN ASM_REWRITE_TAC[]
        THEN PSET_TAC[])
THEN RW_TAC std_ss[]
THEN ASM_REWRITE_TAC[]

THEN `?A1'.( (A:num->'a->bool) (1:num) = PREIMAGE ((X:num->'a->'b) t) A1' INTER p_space p)/\
A1' IN measurable_sets M` by

  (UNDISCH_TAC(``!i.
             i IN {0; 1:num} ==>
             (A:num->'a->bool) i IN
             {PREIMAGE f A INTER p_space p |
             (f = if i = 0 then (X:num->'a->'b) s else X t) /\ A IN measurable_sets M}``)
     THEN DISCH_TAC
     THEN POP_ASSUM (MP_TAC o Q.SPEC `(1:num) `)
     THEN FULL_SIMP_TAC std_ss[prove(``1 IN {0;(1:num)}``,
     SRW_TAC[][EXTENSION,GSPECIFICATION])]
      THEN `A 1 IN
	 {PREIMAGE f A INTER p_space p |
	  (f = X t) /\ A IN measurable_sets M} = A 1 IN
	  {PREIMAGE (X t) A INTER p_space p|
	  A IN measurable_sets M }` by  ASM_SET_TAC[]
         THEN ASM_REWRITE_TAC[]
        THEN PSET_TAC[])

THEN ASM_REWRITE_TAC[]				
THEN UNDISCH_TAC(`` !A':num->'a->bool.(!i'.
                  i' IN {s:num; t} ==>
                  A' i' IN
                  {PREIMAGE ((X:num->'a->'b) i') A INTER p_space p | A IN measurable_sets M}) ==>
             (prob p (BIGINTER {A' j | (j = s) \/ (j = t)}) =
              Normal (product {s; t} (\j. real (prob p (A' j)))))``)
THEN DISCH_TAC
THEN POP_ASSUM (MP_TAC o Q.SPEC `(\i. if i = s then
				    PREIMAGE ((X:num->'a->'b) s) (A':'b->bool) INTER p_space p else
				    PREIMAGE (X t) (A1':'b->bool) INTER p_space p)`)
THEN DISCH_TAC
 THEN KNOW_TAC(``(!i'.
              i' IN {s:num; t} ==>
              (\i.
                   if i = s then PREIMAGE (X s) A' INTER p_space p
                   else PREIMAGE ((X:num->'a->'b) t) A1' INTER p_space p) i' IN
              {PREIMAGE (X i') A INTER p_space p | A IN measurable_sets M}) ``)
      THEN1(ASM_SET_TAC[])
THEN DISCH_TAC
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(``{if j = (s:num) then PREIMAGE (X s) A' INTER p_space p
              else PREIMAGE ((X:num->'a->'b) t) A1' INTER p_space p |
              j |
              (j = s) \/ (j = t)} =
	      {PREIMAGE (X s) (A':'b->bool) INTER p_space p;
	       PREIMAGE (X t) (A1':'b->bool) INTER p_space p}``)
     THEN1(RW_TAC std_ss [GSPECIFICATION, EXTENSION]
           THEN PSET_TAC [PREIMAGE_def]
           THEN EQ_TAC
	      THEN1(RW_TAC std_ss[]
	            THEN1( ASM_SET_TAC[])
	            THEN ASM_SET_TAC[])
	   THEN DISCH_TAC
          THEN RW_TAC std_ss[]
	  THEN1(Q.EXISTS_TAC`s:num`
		 THEN ASM_SET_TAC[])

          THEN Q.EXISTS_TAC`t:num`
	  THEN RW_TAC std_ss[])
THEN DISCH_TAC
THEN FULL_SIMP_TAC std_ss[]
THEN `FINITE {s:num} /\FINITE {t:num} /\ DISJOINT {s:num} {t:num} /\ ( {s:num;t} = {s} UNION {t}) `
  by SRW_TAC[] [GSPECIFICATION, EXTENSION]
THEN  FULL_SIMP_TAC std_ss [PRODUCT_UNION]
THEN FULL_SIMP_TAC std_ss [PRODUCT_SING]
THEN FULL_SIMP_TAC std_ss [BIGINTER_INTER]
THEN FULL_SIMP_TAC std_ss [REAL_MUL_COMM]);



val BIGINTER_3_sets = store_thm("BIGINTER_3_sets",
  `` ! x y z . BIGINTER {x;y;z} = x INTER y INTER z``,
SRW_TAC [] [GSPECIFICATION, EXTENSION, BIGINTER]
THEN EQ_TAC
     THEN1(RW_TAC std_ss[])
THEN ASM_SET_TAC[]);



val prob_prod_any_2_of_10 = store_thm("prob_prod_any_2_of_10",
  ``   !p M X ii A s t.
       s <> t /\ prob_space p /\ indep_vars p M X ii /\
       {s; t} SUBSET ii /\
       (!i. i IN {s; t} ==> A i IN measurable_sets (M i)) ==>
       (prob p
          (PREIMAGE (X s) (A s) INTER p_space p INTER
           (PREIMAGE (X t) (A t) INTER p_space p)) =
        prob p (PREIMAGE (X s) (A s) INTER p_space p) *
        prob p (PREIMAGE (X t) (A t) INTER p_space p))``,

RW_TAC std_ss[]
THEN KNOW_TAC(``FINITE {s:num;t}``)
     THEN1 (SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF])
THEN DISCH_TAC
THEN KNOW_TAC(``prob p (BIGINTER {PREIMAGE (X i) (A i) INTER p_space p | i IN {s; t}}) =
     		 Normal
		   (product {s:num; t}
		        (\i. real (prob p (PREIMAGE (X i) (A i) INTER p_space p)))) ``)

      THEN1(DEP_REWRITE_TAC [indep_varsD]
            THEN Q.EXISTS_TAC `M`
	    THEN Q.EXISTS_TAC `ii`
	    THEN RW_TAC std_ss[]
	    THEN SRW_TAC[] [GSPECIFICATION, EXTENSION])
THEN DISCH_TAC
THEN KNOW_TAC(``{PREIMAGE ((X:num->'a->'b) i) (A i) INTER p_space p | i IN {s:num; t}} =
                 {PREIMAGE (X s) (A s) INTER p_space p ;
		  PREIMAGE (X t) (A t) INTER p_space p}``)
     THEN1 (ASM_SET_TAC[])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(``BIGINTER
                   {PREIMAGE ((X:num->'a->'b) s) (A s) INTER p_space p;
                    PREIMAGE (X t) (A t) INTER p_space p} =
		 (PREIMAGE (X s) (A s) INTER p_space p) INTER
                 (PREIMAGE (X t) (A t) INTER p_space p )``)
      THEN1(RW_TAC std_ss [BIGINTER_INTER])
THEN DISCH_TAC
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(``FINITE {s:num} /\ FINITE {t:num} /\
                 DISJOINT {s:num} {t:num} /\ ({s:num;t} = {s} UNION {t})``)

     THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF, DISJOINT_DEF])
THEN DISCH_TAC
THEN RW_TAC std_ss [PRODUCT_UNION]
THEN RW_TAC std_ss [PRODUCT_SING]
THEN KNOW_TAC(``random_variable ((X:num->'a->'b) s) p (m_space (M s), measurable_sets (M s)) /\
                 random_variable  ((X:num->'a->'b) t)p (m_space (M t), measurable_sets (M t))``)
     THEN1(FULL_SIMP_TAC std_ss [indep_vars]
           THEN UNDISCH_TAC(`` !i:num.
           			   i IN ii ==>
                            random_variable ((X:num->'a->'b) i) p (m_space (M i),
			            measurable_sets (M i))``)

	   THEN DISCH_TAC
	   THEN RW_TAC std_ss[]
	        THEN1 (POP_ASSUM (MP_TAC o Q.SPEC `s:num `)
	               THEN ASM_SET_TAC[])

           THEN POP_ASSUM (MP_TAC o Q.SPEC `t:num `)
	   THEN ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``PREIMAGE ((X:num->'a->'b) s) (A s) INTER p_space p IN events p/\
                 PREIMAGE ((X:num->'a->'b) t) (A t) INTER p_space p IN events p``)
     THEN1( REWRITE_TAC [events_def]
            THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``)
                 THEN1(RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
            THEN DISCH_TAC
            THEN ONCE_ASM_REWRITE_TAC[]
            THEN REWRITE_TAC [p_space_def, m_space_def]
            THEN KNOW_TAC(`` (A:num->'b->bool) s IN measurable_sets (M s) /\
                              (A:num->'b->bool) t IN measurable_sets (M t)``)
                 THEN1(RW_TAC std_ss[]
 		       THEN1(UNDISCH_TAC(`` !i:num. i IN {s; t} ==>
		               (A:num->'b->bool) i IN measurable_sets (M i)``)
			     THEN DISCH_TAC
			     THEN POP_ASSUM (MP_TAC o Q.SPEC `s:num `)
			     THEN ASM_SET_TAC[])
		THEN UNDISCH_TAC(`` !i:num. i IN {s; t} ==>
		                     (A:num->'b->bool) i IN measurable_sets (M i)``)
                THEN DISCH_TAC
		THEN POP_ASSUM (MP_TAC o Q.SPEC `t:num `)
		THEN ASM_SET_TAC[])
           THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def])
THEN DISCH_TAC
THEN RW_TAC std_ss [normal_real_mul, PROB_FINITE]
THEN RW_TAC std_ss [normal_real, PROB_FINITE]);


val _ = export_theory();













